#ifndef files_h
  #define files_h

#define UNICODE
#define _UNICODE

#include <sys/types.h>
#ifdef __linux__
  #include <linux/limits.h>
  #include <sys/ipc.h>
  #include <sys/sem.h>
  #include <sys/msg.h>
  #include <sched.h>
  #include <sys/socket.h>
  #include <netinet/in.h>
  #include <sys/mman.h>
  #include <arpa/inet.h>
  #include <sys/ioctl.h>
  #include <sys/shm.h>
#elif defined(_WIN32) || defined(_WIN64)
  #ifndef NAME_MAX
    #define NAME_MAX FILENAME_MAX
  #endif
#endif
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/file.h>

#ifdef __linux__
  #include <sys/select.h>
#elif defined(_WIN32) || defined(_WIN64)
  #include <winsock2.h>
#endif

#include "errcodes.h"
#include "semaphores.h"
#include "messages.h"
#include "filer.h"
#include "global.h"
#include "md5.h"
#include "prunit.h"

#define DefaultPort             7777
#define DefaultHTTPPort         7780

#define mode_Server             1
#define mode_Client             0

#define init_MaxFilesServer	    4096
#define init_MaxFilesClient	    256
#define init_MaxKeys            256
#define default_LockSegLevel	14

#define filetype_Data			0
#define filetype_RAW			1
#define filetype_Ini			2
#define filetype_Txt			3
#define filetype_TempDat        4
#define filetype_LocalTempData  5

#define TXTBufferSize			65536
#define CopyBufferSize			65536
#define StrBufferSize			256

#define DefaultPreloadSize      16384

#define PacketDataSize          1472

#define rm_Flags_Preload        0x0001
#define rm_Flags_WaitMessage    0x0002
#define rm_Flags_Progress       0x0004
#define rm_Flags_File           0x0008
#define rm_Flags_MD5            0x0010

#define fileflags_Open          0x0001
#define fileflags_ReadDone      0x0002
#define fileflags_DataLocked    0x0004
#define fileflags_IndexLocked   0x0008


#define KeyNameSize             33

#include <stdint.h>

#include "filer.h"

enum extensions {
    ext_none=0,
    ext_dat,
    ext_inx,
    ext_tmp,
    ext_sav
};

extern int MaxFiles;
extern char pdsDir[];
extern char LogName[];
extern char DefaultKey[]; // mel by byt vzdy delsi nez 8 znaku, at nekoliduje s pripadnym SubKey
extern char datExtension[];	// pozor !!! delka datExtension a inxExtension musi byt vzdy shodna
extern char inxExtension[];
extern char tmpExtension[];
extern char savExtension[];
extern uint8_t LockSegLevel;
extern char LogStamp[80];
extern int PreloadSize;
extern int PreloadEnabled;
extern char ConfName[];
extern char PassFileName[];
extern char backupExtension[];
extern char pdsID[];
extern time_t start_time;
extern uint8_t CheckIntegrity;
extern mode_t NewFileMode;

typedef struct pds_lock pds_lock,*P_pds_lock;

struct pds_lock {
	int				Sector;
	int 			Locked;		// pocet uzamceni v LockTab
	P_pds_lock		NextLock;
	uint8_t			Locks[];
};

typedef struct page_buffer page_buffer,*P_page_buffer;

struct page_buffer {
	page_buffer *		Next;
	int					PRef;
	int					Size;
	unsigned short		Key;
	uint8_t				Valid;
	uint8_t				Dummy; // vyplnovy znak na zarovnani delky na nasobek 4
};

typedef struct pds_key pds_key,*P_pds_key;

struct pds_key {
    char                Key[KeyNameSize];
    int                 UseCount;
    pid_t               Locker;
    int                 LockCount;
    unsigned char       Password_MD5[16];
    uint8_t             ValidPassword;
};

typedef struct pds_file pds_file,*P_pds_file;

struct pds_file {
	unsigned short		Random;  	// nahodna hodnota
	unsigned short		Flags;   	// bitove priznaky 0x0001 - soubor otevren
									//				   0x0002 - pro TXT a INI soubory pomocny priznak "vsechna data nactena"
									//                 0x0004 - datovy soubor uzamcen flock();
									//                 0x0008 - indexovy soubor uzamcen flock();
	unsigned short		FileType;   // typ souboru 0 - Data, 1 - RAW, 2 - Ini, 3 - Txt, 4 - Temp
	unsigned short		FileGen; 	// pri otevreni se dosazuje hodnota z Generation++
	int 				Handle;
	IsamSmallInfoRec	datInfo;
	int					InxHandle;
	int					BlockLen;
	int					MaxPages;
	int 				NeedIndexBufferSize;
	IsamDatIndDescr		InxDescr[MaxNrOfKeys+1];
	char *				TXTBuffer;
	int		            TXTBufferPos;
	int		            InTXTBuffer;
	int					TXTFilePos;
	int					NextFree;
	int					LockLevel;
	pid_t 				Locker;	 	// identifikace procesu, ktery si soubor vyhradil pro sebe
	int					ReadLockLevel;
	pid_t 				ReadLocker;	// identifikace procesu, ktery si soubor vyhradil pro sebe
	pid_t				Approp;  	// identifikace procesu, ktery aktualne pracuje se souborem
	int 				AppropLevel;// pocet vyhrazeni jednim procesem
	pid_t				Expectant;	// identifikace procesu, ktery bude Approp po dokonceni procesu aktualne zpracujiciho soubor
	int 				Counter; 	// pocet otevreni, pri otevreni++, pri uzavreni--, pokud po uzavreni 0, tak se uvolnuje
	int 				LockSegMask;
	int					LockSegCount;
	P_pds_lock			FirstLock;
	page_buffer	*		PageBuffers;
	uint8_t				LockSegLevel;
	uint8_t				LockSectShift;
	uint8_t				Semaphores[32];
	char				Key[KeyNameSize];
	char				SubKey[9];
	char				Name[13];
	void *              PreloadBuffer;
	int                 PreloadBufferSize;
	int                 PreloadFree;
    void *              PreloadBufferPos;
    P_PreloadHeader     PreloadHead;
    uint8_t             PreloadValid;
    char *              RealName;
    char *              RealIndexName;
};

typedef struct pds_files pds_files,*P_pds_files;

struct pds_files {
	int		            MaxFiles;
	int	                Opened;
	int	                FirstFree;
	int	                FreeFiles;
	pds_key *           Keys;
	int                 MaxKeys;
	int                 ActKeys;
	int                 UseCount;
	unsigned short      Generation; // pouze 16bit z duvodu, ze identifikace souboru klientem je 64 bit = 16bit FileGen+16bit Random+32 bit pozice v poli
	unsigned long int   TotalMessages;
	struct pds_file     Files[];
};

extern int * _Files_Sorted;

extern pds_files * _Files;
//* extern int shm_Files;
//* extern int shm_Files_Sorted;

extern int _FilesSize;
extern int SortedSize;

typedef struct fileID fileID,*P_fileID;

#pragma pack(push,1)

extern fileID NullFileID;

typedef struct ArchiveHead ArchiveHead,*P_ArchiveHead;
struct ArchiveHead {
	uint8_t	    StartOfHead;
	uint32_t    Count;
	uint8_t     MD5[16];
	uint32_t    Size;
	uint32_t    PackedSize;
	uint8_t     NameLength;
};

typedef struct IntArr IntArr,*P_IntArr;
struct IntArr {
    int32_t     int0;
    int32_t     int1;
};

#pragma pack(pop)

extern IsamInfoRec	dummy_IR;

void ClearPdsFile(pds_file * pds);
char * Stamp();
int InitFiles(int Mode);
void DoneFiles();

typedef struct local_files local_files,*P_local_files;
typedef struct pds_child pds_child,*P_pds_child;

struct pds_child {
	pid_t 			pid;
	time_t			LastActive;
	P_local_files	Local;
    P_pds_child		NextChild;
	P_pds_child		PrevChild;
};

typedef struct IndexPertains IndexPertains,*P_IndexPertains;

struct IndexPertains {
	uint8_t			SequentialOK;
	uint8_t			PathInd;
	IsamSearchStep	Path[MaxHeight];
	int				LastInvalidator;
    uint8_t         IsValidLast;
	IsamKeyStr		LastSearchKey;
	int 			LastSearchRefNr;
};

typedef struct LockItem LockItem,*P_LockItem;

struct LockItem {
	int 			Counter;
	int 			RefNr;
};

typedef struct local_file local_file,*P_local_file;

struct local_file {
	int					Index;
	int					NextItem;
	int					PrevItem;
	IndexPertains		IndexArr[MaxNrOfKeys];
	int					ActLocks;
	int					MaxLocks;
	int					FilePos; // platne pouze pro soubor typu RAW
	uint8_t				Semaphores[32];
	uint8_t             PrevPreCount;
	uint8_t             NextPreCount;
    P_local_files       MyFiles;
	LockItem *			MyLocks;
};

struct local_files {
	int					_Local_size;
	int			 		max_local_files;
	char *		 		local_buffer;
	void *				IndexBuffer;
	int					IndexBufferSize;
	int 				Opened;
	int					FirstFree;
	int					FirstOpen;
	int					client_sockfd;
	pid_t				child_pid;
	char *				IdentStr;
	unsigned short		IdentStrSize;
	unsigned short		IdentStrCount;
	const char *		FuncName;
	P_pds_child			SelfChild;
	unsigned short      LastErrorCode;
	uint8_t             DisableCheckFreeze;
	int                 ActCounter;
	time_t              ProgressTime;
	unsigned long long  ActProgress;
	unsigned long long  TotalProgress;
	int                 MyKey;                  /* Ukazatel na polozku pole pds_files.Keys */
	int                 AuthorizedKey;
    fd_set              clientfds;
    uint8_t *           LiveSocket;
	uint8_t             PrevPreCount;
	uint8_t             NextPreCount;
	char                LogBuffer[StrBufferSize];
	unsigned short int  LastMessage;
	long unsigned int   Messages;
    union {
        uint32_t        ip;
        uint8_t         ip8[4];
    };
    unsigned short      port;
    time_t              start;
	local_file			Files[];
};

extern const int local_buffer_size;

/*
P_local_files	_Local=NULL;
int				_Local_size;
int			 	max_local_files=0;
const int		local_buffer_size=65536;
char *		 	local_buffer=NULL;
void *			IndexBuffer=NULL;
int				IndexBufferSize=0;
*/

int GetLastFileError(P_local_files Local,int AltError);
void SetLastFileError(P_local_files Local,int ErrNum,int AltError);
void CloseClient(int * client_sockfd);
int CheckLocalIntegrity(P_local_files _Local);
int DoubleLocalFiles(P_local_files * Local);
void CloseAllLocalFiles(P_local_files _Local);
void LocalDoneFiles(P_local_files * Local);
int LocalInitFiles(P_local_files * Local);
int FindFile(char *Key,char *SubKey,char *FileName,unsigned short FileType,int *Index);
int AppropriateFile(P_local_files _Local,fileID fID,unsigned short Type,uint8_t ForWrite);
int UnAppropriateFile(fileID fID);
void FreePageBuffers(P_pds_file File);
void CloseGlobalFile(int Index,uint8_t LockSemaphore);
int FindLock(P_local_file File,int RefNr,int *Index);
P_pds_lock GetLockSector(P_local_files Local,P_pds_file pds,int sector,uint8_t Create);
int UnlockRec(P_local_file File,int RefNr,unsigned short All);
int UnlockAllRecs(P_local_file File);
P_page_buffer GetPageBuffer(P_local_file File,unsigned short Key,int PRef);
void CloseLocalFile(P_local_files _Local,int Index);

extern const char ValidFileNameChars[];

int IsValidFileName(char * FileName);
int ValidateFileName(char * FileName);
int IsamPutBlock(P_local_files Local,int Handle,int Offset,unsigned short int Length,void * Source);
int IsamPutDummyBlock(P_local_files Local,int Handle,int Ref,unsigned short Len);
int IsamCopyInfoRecToIR(P_pds_file File,unsigned short KeyNr,IsamInfoRec * IR);
int IsamPutInfoRec(P_local_files Local,P_pds_file File,unsigned short KeyNr,uint8_t IndFDamaged);
int IsamInitDatFile(P_local_files Local,int Handle,unsigned short RecLen,unsigned short NrOfKeys);
int CreateInxFileStructure(P_local_files Local,P_pds_file File);

int OpenDat(P_local_files Local,P_pds_file file, char * name);
int OpenInx(P_local_files Local,P_pds_file file, char * name);

typedef struct InitNewInxFile InitNewInxFile,*P_InitNewInxFile;

struct InitNewInxFile {
	uint8_t 	KeyLen;
	uint8_t 	AllowDupKeys;
	uint8_t 	XlatKeys;
	uint8_t		Dummy; // nepouzito, je zde pro zarovnani jednobajtovyh parametru na hranici dvouslova
	uint16_t 	CodePage;
	uint16_t	IndexVersion;
};

typedef struct InitNewDatFile InitNewDatFile,*P_InitNewDatFile;

struct InitNewDatFile {
	unsigned short RecSize;
	unsigned short NrOfKeys;
	InitNewInxFile Indexs[];
};

typedef struct DirItem DirItem,*P_DirItem;
struct DirItem {
    char    Name[NAME_MAX];
    int     Size;
    uint8_t UseFlags;
};

int     IsamInitInxFile(P_local_files Local,P_pds_file File,P_InitNewDatFile NewInfo);
int     ExistFile(P_local_files Local,char *Key,char *SubKey,char *FileName,uint8_t * Exist,const char * Path);
int     FileSize(P_local_files Local,char *Key,char *SubKey,char *FileName,int * Size,const char * Path);
int     GetFileInfo(P_local_files Local,char *Key,char *SubKey,char *FileName,FileInfoRec * FileInfo,const char * Path);
int     AllocLocalHandle(P_local_files * Local,int Index);
int     OpenFile(P_local_files * Local,char *Key,char *SubKey,char *FileName,unsigned short FileType,uint8_t NewAllowed,uint8_t Truncate,P_InitNewDatFile NewInfo,const char * Path);
int     OpenFileDirect(P_local_files * Local,const char * _RealName,const char * _FileName,unsigned short FileType,uint8_t NewAllowed,uint8_t Truncate,uint8_t ReadOnly);
int     CheckFileID(fileID fID);
int     CheckLocalFileID(P_local_files _Local,fileID fID,unsigned short int msgID);
int     CloneFile(P_local_files * Local,fileID fID,unsigned short fileType);
int     LocalIndexToGlobal(P_local_files _Local,P_fileID fID,unsigned short int msgID);
int     AppropriateLocalFile(P_local_files _Local,fileID fID,unsigned short Type,uint8_t ForWrite,unsigned short int msgID);
int     UnAppropriateLocalFile(P_local_files _Local,fileID fID);
int     LockRec(P_local_files _Local,fileID fID,int RefNr);
int     RecIsLocked(P_local_file File,int RefNr,uint8_t * Result);
char *  GetRealName(P_local_files Local,char *Key,char *SubKey,char *FileName,const char * Path,extensions Ext=ext_none);
int     CreateSubKey(P_local_files Local,char *Key,char *SubKey,const char * Path);
int     IsTrivialError(unsigned short Error);
int     _send(int client_sockfd,void * Data,int Size);
int     Send(int client_sockfd,void * Data,int Size);
int     SendOK(int client_sockfd,int SendCounter);
int     SendWait(int client_sockfd,int SendCounter,unsigned short Progress);
int     SendFile(P_local_files Local,int SendCounter,int FileHandle);
int     LockKey(P_local_files Local,char *Key,const char * Path);
int     UnlockKey(P_local_files Local,char *Key,const char * Path);
int     TestArchive(P_local_files Local,char * Key, char * SubKey,int Archive,const char * Path);
int     ExtractArchive(P_local_files Local,char * Key, char * SubKey,int Archive,const char * Path);
int     GetKeyPassword(P_local_files Local,char *Key,const char * Path);
int     CompareKeyPassword(P_local_files Local,char *Key,void * MD5,const char * Path);
int     GetRec(P_local_files _Local,fileID fID,int RefNr,void * Buffer,int BufferSize,uint8_t IgnoreLock);
int     PutRec(P_local_files _Local,fileID fID,int RefNr,void * Buffer);
int     ResetIndex(P_local_files Local,P_pds_file File,P_InitNewDatFile NewInfo);
int     GetBlock(P_local_files _Local,fileID fID,unsigned int Offset,unsigned int Size,void * Dest);
int     PutBlock(P_local_files _Local,fileID fID,unsigned int Offset,unsigned int Size,const void * Src);
int     UnLockIndex(P_pds_file file);
int     CloseIndex(P_pds_file file);
int     OpenFileExclusive(const char * Name);
void    CloseFileExclusive(int Handle);
fileID  GetLocalFileID(P_local_files _Local,int Index);
fileID  GetFileID(P_local_files _Local,int Index);
char *  GetFileName(P_local_files _Local,int Index);

void    InitLogBuffer(P_local_files Local);
void    AddToLogBuffer(P_local_files Local,int level,const char* format, ... );
void    InitStrBuffer(char * StrBuffer);
void    AddToStrBuffer(char * StrBuffer,int level,const char* format, ... );

#include "sixpack.h"

#endif
