#include "wincompat.h"

#include <sys/types.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>
#include <winsock2.h>
#include <windows.h>

static const char filename_letters[]="abcdefghijklmnopqrstuvwxyz0123456789!#$%&'()-@^_`{}~";

int mkstemp(char *tmpl)
{
  int len;
  char *XXXXXX;
  static unsigned long long value;
  unsigned long long random_time_bits;
  unsigned int count;
  int fd = -1;
  int save_errno = errno;

#define ATTEMPTS_MIN (52 * 52 * 52)
#if ATTEMPTS_MIN < TMP_MAX
  unsigned int attempts = TMP_MAX;
#else
  unsigned int attempts = ATTEMPTS_MIN;
#endif

  len = strlen (tmpl);
  if (len < 6 || strcmp (&tmpl[len - 6], "XXXXXX"))
    {
      errno = EINVAL;
      return -1;
    }

  XXXXXX = &tmpl[len - 6];

  {
    SYSTEMTIME stNow;
    FILETIME ftNow;

    // get system time
    GetSystemTime(&stNow);
    stNow.wMilliseconds = 500;
    if (!SystemTimeToFileTime(&stNow, &ftNow))
    {
        errno = -1;
        return -1;
    }

    random_time_bits = (((unsigned long long)ftNow.dwHighDateTime << 32)
                        | (unsigned long long)ftNow.dwLowDateTime);
  }
  value += random_time_bits ^ (unsigned long long)GetCurrentThreadId();

  for (count = 0; count < attempts; value += 7777, ++count)
  {
    unsigned long long v = value;

    XXXXXX[0] = filename_letters[v % 52];
    v /= 52;
    XXXXXX[1] = filename_letters[v % 52];
    v /= 52;
    XXXXXX[2] = filename_letters[v % 52];
    v /= 52;
    XXXXXX[3] = filename_letters[v % 52];
    v /= 52;
    XXXXXX[4] = filename_letters[v % 52];
    v /= 52;
    XXXXXX[5] = filename_letters[v % 52];

    fd = open (tmpl,O_RDWR|O_CREAT|O_EXCL|O_BINARY,S_IRUSR|S_IWUSR);
    if (fd >= 0)
    {
      errno = save_errno;
      return fd;
    }
    else if (errno != EEXIST)
    return -1;
  }

  errno = EEXIST;
  return -1;
}

void syslog(int priority, const char *format, ...)
{
    va_list args;
    va_start(args,format);
    printf(format,args);
    va_end(args);
}
