#ifndef childs_h
	#define childs_h

#include <errno.h>
#include <sys/types.h>
#include <limits.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <unistd.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/mman.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <stdint.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/shm.h>
#include <linux/sched.h>
#include <time.h>
#include <fnmatch.h>
#include <dirent.h>
#include <syslog.h>

#include "version.h"
#include "prunit.h"
#include "errcodes.h"
#include "messages.h"
#include "semaphores.h"
#include "files.h"
#include "filer_proc.h"
#include "ini.h"
#include "global.h"
#include "md5.h"
#include "sixpack.h"
#include "http.h"

#ifndef CHILD_MAX
	#define CHILD_MAX		999
#endif

#define bufsize				8192
#define stack_size			65536
#define MinDirItems         1024    /* Melo by byt 1 nebo 2^ */
#define ParamBufSize        65534

#define TrivialErrors		        14

#define preload_NextKey              1
#define preload_PrevKey              2
#define preload_FindKey              3
#define preload_SearchKey            4
#define preload_FindKeyAndRef        5

#define changefile_Dat               1
#define changefile_Inx               2
#define changefile_Packed         0x80

#define max_Preload_Items           16

extern int MaxOut;
extern int TimeOut;
extern unsigned short TrivialErrorList[TrivialErrors];

extern char compilation_time[];
extern char compilation_date[];

typedef struct pds_childs pds_childs,*P_pds_childs;

struct pds_childs {
	int 		MaxChilds;
	int 		ChildsCount;
	int 		FreeChilds;
	int			BusyChilds;
	P_pds_child FirstFree;
	P_pds_child Unused;
	P_pds_child Busy;
	pds_child	childs[];
};

extern pds_childs * _Childs;

typedef struct pds_accesslist pds_accesslist,*P_pds_accesslist;

struct pds_accesslist {
    union {
        uint32_t    ip;
        uint8_t     ip8[4];
    };
    uint8_t         mask;
    uint8_t         allow;
};

typedef struct pds_params pds_params,*P_pds_params;

struct pds_params {
    local_files     *   Local;
    short int           port;
    char            *   path;
    unsigned int        masks[33];
    int                 access_count;
    pds_accesslist  *   access;
    mode_t              NewFileMode;
    char            *   LogName;
    uint8_t             LogMode;
    uint8_t             CheckIntegrity;
    char                Name[33];
    uint8_t             Echo;
    uint8_t             Http;
    short int           httpport;
};

extern pds_params * Params;

/*******/

P_pds_child AllocChild();
void ChildsStatus();
int InitChilds();
pid_t GetFreeChildpid();
int CountStrings(char * buffer,int size);
void InitResultHead(P_ResultMsg ResMsg,int counter);
int IsTrivialError(unsigned short Error);
int SendError(int client_sockfd,unsigned short ErrNum,int SendCounter,const char * FuncName);
int ChildAction(P_child_message msg,void * buffer,int size,P_local_files * Local);
void PrintBuffer(P_local_files Local,char * Buffer,int Size);
P_pds_child SearchChild(pid_t ChildPid);
int ChildProc(void *data);
int DeleteChild(pid_t pid);
int FreeChild(pid_t pid);
void EmployChild(P_pds_child child);
void DoneChilds();
void WorkMsg(status_message msg);
void DoneParams();
int InitParams();
int GetParams();
void OpenLog();

#endif
