#include "filer_proc.h"

int KeyNr=-1;

void CopyKeyStr(IsamKeyStr * ToKey,const IsamKeyStr * FromKey)
{
	ToKey->Length=FromKey->Length;
	if(ToKey->Length>MaxKeyLen)
		ToKey->Length=MaxKeyLen;
	if(ToKey->Length)
	{
		memcpy(&ToKey->Data,&FromKey->Data,ToKey->Length);
	}
}

int CmpKeyStr(const IsamKeyStr& Key1,const IsamKeyStr& Key2)
{
	uint8_t Len=Key1.Length;
	int Result=0;
	if(Len>Key2.Length)
		Len=Key2.Length;
	if(Len)
	{
		Result=memcmp(&Key1.Data,&Key2.Data,Len);
	}
	if(Result==0)
	{
		Result=(int)Key1.Length-(int)Key2.Length;
	}
	if(Result>0)
		Result=1;
	else
	if(Result<0)
		Result=-1;
	return Result;
}

IsamPagePtr GetIsamPagePtrFromBuffer(P_page_buffer PageBuffer)
{
	IsamPagePtr PagePtr=(IsamPagePtr)PageBuffer;
	char * LocPCh=(char *)PagePtr;
	LocPCh+=sizeof(page_buffer);
	PagePtr=(IsamPagePtr)LocPCh;
	return PagePtr;
}

P_page_buffer GetPageBufferFromIsamPagePtr(IsamPagePtr IPP)
{
	P_page_buffer PageBuffer=(P_page_buffer)IPP;
	char * LocPCh=(char *)PageBuffer;
	LocPCh-=sizeof(page_buffer);
	PageBuffer=(P_page_buffer)LocPCh;
	return PageBuffer;
}

void IsamUnPack(IsamPagePtr Page, unsigned short KeyL)
{
	unsigned short I,K,S;
	uint8_t * P=(uint8_t *)Page;
	K=KeyL+9;
	S=(Page->ItemsOnPage-1)*K+6;
	if(KeyL!=MaxKeyLen)
	{
		for(I=Page->ItemsOnPage;I>=2;I--)
		{
			memmove(&Page->ItemArray[I-1],&P[S],K);
			S-=K;
		}
	}
}

void IsamPack(IsamPagePtr Page, unsigned short KeyL)
{
	unsigned short I,K,S;
;
	uint8_t * P=(uint8_t *)Page;
	K=KeyL+9;
	S=K+6;
	if(KeyL!=MaxKeyLen)
	{
		for(I=1;I<Page->ItemsOnPage;I++)
		{
			memmove(&P[S],&Page->ItemArray[I],K);
			S+=K;
		}
	}
}

int IsamGetPage(P_local_file File,int PRef,unsigned short Key,IsamPagePtr * IPgPtr)
{
	P_page_buffer PageBuffer=GetPageBuffer(File,Key,PRef);
	if(PageBuffer==NULL)
	{
		return 0;
	}
	*IPgPtr=GetIsamPagePtrFromBuffer(PageBuffer);
	if(!PageBuffer->Valid)
	{
		int LocI=PRef*_Files->Files[File->Index].BlockLen+_Files->Files[File->Index].InxDescr[Key].BlockOfs;
		if(lseek(_Files->Files[File->Index].InxHandle,LocI,SEEK_SET)!=LocI)
		{
			File->MyFiles->LastErrorCode=errno;
			if(!File->MyFiles->LastErrorCode)
				File->MyFiles->LastErrorCode=err_SeekError;
			return 0;
		}
		else
		if(read(_Files->Files[File->Index].InxHandle,(*IPgPtr),_Files->Files[File->Index].InxDescr[Key].LenRec)!=_Files->Files[File->Index].InxDescr[Key].LenRec)
		{
			File->MyFiles->LastErrorCode=errno;
			if(!File->MyFiles->LastErrorCode)
				File->MyFiles->LastErrorCode=err_IsamReadError;
			return 0;
		}
		IsamUnPack(*IPgPtr,_Files->Files[File->Index].InxDescr[Key].KeyLen);
		PageBuffer->Valid=1;
	}
	return 1;
}

int IsamWritePage(P_local_file File,P_page_buffer PageBuffer)
{
	if((!PageBuffer->Valid)||(PageBuffer->Key==0)||(PageBuffer->Key>_Files->Files[File->Index].datInfo.NrOfKeys))
	{
		File->MyFiles->LastErrorCode=err_InvalidBuffer;
		return 0;
	}
	if((unsigned short)PageBuffer->Size!=sizeof(IsamPage)+sizeof(page_buffer))
	{
		File->MyFiles->LastErrorCode=err_InvalidBuffer;
		return 0;
	}
	IsamPagePtr Buffer=GetIsamPagePtrFromBuffer(PageBuffer);
	int LocI=PageBuffer->PRef*_Files->Files[File->Index].BlockLen+_Files->Files[File->Index].InxDescr[PageBuffer->Key].BlockOfs;

	_Files->Files[File->Index].InxDescr[PageBuffer->Key].Invalidator++;

	if(lseek(_Files->Files[File->Index].InxHandle,LocI,SEEK_SET)!=LocI)
		{
			File->MyFiles->LastErrorCode=errno;
			if(!File->MyFiles->LastErrorCode)
				File->MyFiles->LastErrorCode=err_SeekError;
			return 0;
		}
		else
		{
			IsamPack(Buffer,_Files->Files[File->Index].InxDescr[PageBuffer->Key].KeyLen);
			if(write(_Files->Files[File->Index].InxHandle,Buffer,_Files->Files[File->Index].InxDescr[PageBuffer->Key].LenRec)!=_Files->Files[File->Index].InxDescr[PageBuffer->Key].LenRec)
			{
				IsamUnPack(Buffer,_Files->Files[File->Index].InxDescr[PageBuffer->Key].KeyLen);
				File->MyFiles->LastErrorCode=errno;
				if(!File->MyFiles->LastErrorCode)
					File->MyFiles->LastErrorCode=err_IsamWriteError;
				return 0;
			}
			IsamUnPack(Buffer,_Files->Files[File->Index].InxDescr[PageBuffer->Key].KeyLen);
		}
	return 1;
}

int IsamCompKeys(struct IsamKeyStr Key1,struct IsamKeyStr Key2,int DatRef1,int DatRef2,uint8_t Dup,uint8_t XLat)
{
	int C=0;
	uint8_t Len=Key1.Length;
	if(Key2.Length<Len)
		Len=Key2.Length;
	if(XLat)
	{
		XTrans(&Key1);XTrans(&Key2);
	}
	C=memcmp(&Key1.Data,&Key2.Data,Len);
	if(C==0)
	{
		if(Key1.Length==Key2.Length)
		{
			if(Dup)
			{
				C=DatRef1-DatRef2;
			}
		}
		else
		{
			C=(int)Key1.Length-(int)Key2.Length;
		}
	}
	if(C<0)
		C=-1;
	else
	if(C>0)
		C=1;
	return C;
}

void ValidateIndex(P_local_file File,unsigned short Key)
{
	File->IndexArr[Key].SequentialOK=1;
	File->IndexArr[Key].LastInvalidator=_Files->Files[File->Index].InxDescr[Key].Invalidator;
}

int IsValidIndex(P_local_file File,unsigned short Key)
{
	int Result=0;
	if(File->IndexArr[Key].PathInd==0)
	{
	    Result=1;
	}
	if(File->IndexArr[Key].SequentialOK)
	{
		if(File->IndexArr[Key].LastInvalidator==_Files->Files[File->Index].InxDescr[Key].Invalidator)
		{
			Result=1;
		}
	}
	return Result;
}

void SetLastKey(IndexPertains * LInx,int * UserDatRef,IsamKeyStr * UserKey)
{
    LInx->IsValidLast=1;
    LInx->LastSearchRefNr=*UserDatRef;
    LInx->LastSearchKey=*UserKey;
}

int RecallLastKey(IndexPertains * LInx,int * UserDatRef,IsamKeyStr * UserKey)
{
    if(LInx->IsValidLast)
    {
        *UserKey=LInx->LastSearchKey;
		*UserDatRef=LInx->LastSearchRefNr;
    }
    return 0;
}

void InvalidateLastKey(IndexPertains * LInx)
{
    LInx->IsValidLast=0;
}

int IsamClearKey(P_local_file File,unsigned short Key)
{
	if((Key>0)&&(Key<=_Files->Files[File->Index].datInfo.NrOfKeys))
	{
		File->IndexArr[Key].PathInd=0;
		InvalidateLastKey(&File->IndexArr[Key]);
		ValidateIndex(File,Key);
	}
	else
	{
		File->MyFiles->LastErrorCode=err_BadKey;
		return 0;
	}
	return 1;
}

int _IsamFindKey(P_local_file File,unsigned short Key,int * UserDatRef,IsamKeyStr * UserKey)
{
	KeyNr=Key;
	int OK=0;
	File->MyFiles->LastErrorCode=0;
	if(File->Index==-1)
	{
		File->MyFiles->LastErrorCode=err_BadDescriptor;
		return 0;
	}
	else
	if((Key>0)&&(Key<=_Files->Files[File->Index].datInfo.NrOfKeys))
	{
		int RefNr;
		IsamPagePtr IPgPtr;
		short C,K,L,R;
		IndexPertains * LInx=&File->IndexArr[Key];
		IsamDatIndDescr * GInx=&_Files->Files[File->Index].InxDescr[Key];

		LInx->SequentialOK=0;
		OK=0;
		LInx->PathInd=0;
		RefNr=GInx->RootRef;
		while((RefNr!=0)&&(!OK))
		{
			LInx->PathInd++;
			LInx->Path[LInx->PathInd-1].PageRef=RefNr;
			if(!IsamGetPage(File,RefNr,Key,&IPgPtr))
				return 0;
			L=1;
			R=IPgPtr->ItemsOnPage;
			do {
				K=(L+R)>>1;
				C=IsamCompKeys(*UserKey,IPgPtr->ItemArray[K-1].KeyStr,*UserDatRef,IPgPtr->ItemArray[K-1].DataRef,
					GInx->AllowDupKeys,GInx->XlatKeys);
				if(C<=0)
					R=K-1;
				if(C>=0)
					L=K+1;
			} while(R>=L);
			if(L-R>1)
			{
				*UserDatRef=IPgPtr->ItemArray[K-1].DataRef;
				R=K;
				OK=1;
			}
			else
			{
				if(R==0)
					RefNr=IPgPtr->BckwPageRef;
				else
					RefNr=IPgPtr->ItemArray[R-1].PageRef;
			}
			LInx->Path[LInx->PathInd-1].ItemArrInd=R;
		}
		if(OK)
		{
		    SetLastKey(LInx,UserDatRef,UserKey);
		}
		else
		{
			while((LInx->PathInd>0)&&(LInx->Path[LInx->PathInd-1].ItemArrInd==0))
				LInx->PathInd--;
            InvalidateLastKey(LInx);
		}
		LInx->SequentialOK=1;
	}
	else
	{
		File->MyFiles->LastErrorCode=err_BadKey;
		return 0;
	}
	return OK;
}

int IsamNextKey(P_local_file File,unsigned short Key,int * UserDatRef,IsamKeyStr * UserKey)
{
	KeyNr=Key;
	if((Key>0)&&(Key<=_Files->Files[File->Index].datInfo.NrOfKeys))
	{
		int RefNr;
		IsamPagePtr IPgPtr=NULL;
		IndexPertains * LInx=&File->IndexArr[Key];
		IsamDatIndDescr * GInx=&_Files->Files[File->Index].InxDescr[Key];

		if(!IsValidIndex(File,Key))
		{
//			if(GInx->SearchForSequentialEnabled)
			{
                RecallLastKey(LInx,UserDatRef,UserKey);
				if(!_IsamFindKey(File,Key,UserDatRef,UserKey))
				{
					if(File->MyFiles->LastErrorCode)
						return 0;
				}
				ValidateIndex(File,Key);
			}
//			else
//			{
//				File->MyFiles->LastErrorCode=err_NotSequentialAccess;
//				return 0;
//			}
		}
		LInx->SequentialOK=0;
		if(LInx->PathInd==0)
			RefNr=GInx->RootRef;
		else
		{
			if(!IsamGetPage(File,LInx->Path[LInx->PathInd-1].PageRef,Key,&IPgPtr))
				return 0;
			RefNr=IPgPtr->ItemArray[LInx->Path[LInx->PathInd-1].ItemArrInd-1].PageRef;
		}
		while(RefNr!=0)
		{
			LInx->PathInd++;
			LInx->Path[LInx->PathInd-1].PageRef=RefNr;
			LInx->Path[LInx->PathInd-1].ItemArrInd=0;
			if(!IsamGetPage(File,RefNr,Key,&IPgPtr))
				return 0;
			RefNr=IPgPtr->BckwPageRef;
			}
		if(LInx->PathInd)
		{
			while((LInx->PathInd>1)&&(LInx->Path[LInx->PathInd-1].ItemArrInd==IPgPtr->ItemsOnPage))
			{
				LInx->PathInd--;
				if(!IsamGetPage(File,LInx->Path[LInx->PathInd-1].PageRef,Key,&IPgPtr))
					return 0;
			}
			if(LInx->Path[LInx->PathInd-1].ItemArrInd<IPgPtr->ItemsOnPage)
			{
				LInx->Path[LInx->PathInd-1].ItemArrInd++;
				CopyKeyStr(UserKey,&IPgPtr->ItemArray[LInx->Path[LInx->PathInd-1].ItemArrInd-1].KeyStr);
				*UserDatRef=IPgPtr->ItemArray[LInx->Path[LInx->PathInd-1].ItemArrInd-1].DataRef;
			}
			else
			{
				LInx->PathInd=0;
			}
		}
		LInx->SequentialOK=1;
		if(!LInx->PathInd)
		{
			File->MyFiles->LastErrorCode=err_NoLargerKeyFound_;
			return 0;
		}
		SetLastKey(LInx,UserDatRef,UserKey);
	}
	else
	{
		File->MyFiles->LastErrorCode=err_BadKey;
		return 0;
	}
	return 1;
}

int IsamNextDiffKey(P_local_file File,unsigned short Key,int * UserDatRef,IsamKeyStr * UserKey)
{
	KeyNr=Key;
	if((Key>0)&&(Key<=_Files->Files[File->Index].datInfo.NrOfKeys))
	{
		IndexPertains * LInx=&File->IndexArr[Key];
		IsamDatIndDescr * GInx=&_Files->Files[File->Index].InxDescr[Key];

		if(LInx->IsValidLast)
		{
            RecallLastKey(LInx,UserDatRef,UserKey);
		}
		else
        {
            return IsamNextKey(File,Key,UserDatRef,UserKey);
        }
        int LocI3;
        IsamKeyStr SaveKeyStr;
        uint8_t Compenzation=1;
        while(Compenzation)
        {
            Compenzation=0;
            LocI3=GInx->KeyLen;
            if(UserKey->Length<LocI3)
            {
                UserKey->Data[UserKey->Length]=0;
                UserKey->Length++;
            }
            else
            {
                uint8_t Znak=(uint8_t)UserKey->Data[LocI3-1];
                if(Znak!=255)
                {
                    if((GInx->XlatKeys)&&(LocI3>1)&&(((Znak=='H')||(Znak=='h'))&&((UserKey->Data[LocI3-2]=='C')||(UserKey->Data[LocI3-2]=='c'))))
                    {
                        if(UserKey->Data[LocI3-2]=='C')
                          UserKey->Data[LocI3-2]='I'; else UserKey->Data[LocI3-2]='i';
                        UserKey->Data[LocI3-1]=0;
                    }
                    else
                    {
                        UserKey->Data[LocI3-1]++;
                    }
                }
                else
                {
                    while(((uint8_t)UserKey->Data[LocI3-1]==255)&&(LocI3>0))
                    {
                        UserKey->Data[LocI3-1]=0;
                        LocI3--;
                    }
                    if(LocI3==0)
                    {
                        File->MyFiles->LastErrorCode=err_NoLargerKeyFound;
                    }
                    else
                    {
                        UserKey->Data[LocI3-1]++;
                    }
                }
            }
            if(File->MyFiles->LastErrorCode==0)
            {
                CopyKeyStr(&SaveKeyStr,UserKey);
                if(IsamSearchKey(File,Key,UserDatRef,UserKey))
                {
                    IsamKeyStr _tempKey;
                    IsamKeyStr _tempKey2;
                    CopyKeyStr(&_tempKey,UserKey);
                    CopyKeyStr(&_tempKey2,&SaveKeyStr);
                    if(GInx->XlatKeys)
                    {
                        XTrans(&_tempKey);XTrans(&_tempKey2);
                    }

                    if(CmpKeyStr(_tempKey2,_tempKey)>=0)
                    {
                        CopyKeyStr(UserKey,&SaveKeyStr);
                        Compenzation=1;
                    }
                }
                else
                {
                    if((File->MyFiles->LastErrorCode==0)||((int)(File->MyFiles->LastErrorCode/100)==102))
                    {
                        File->MyFiles->LastErrorCode=err_KeyNotFound;
                    }
                    return 0;
                }
            }
            else
            {
                File->MyFiles->LastErrorCode=err_Unspecified;
                return 0;
            }
        }
	}
	else
	{
		File->MyFiles->LastErrorCode=err_BadKey;
		return 0;
	}
	return 1;
}

int IsamFindKeyAndRef(P_local_file File,unsigned short Key,int * UserDatRef,IsamKeyStr * UserKey,short NotFoundSearchDirection)
{
	if(!_IsamFindKey(File,Key,UserDatRef,UserKey))
	{
		if(File->MyFiles->LastErrorCode==0)
		{
			if(NotFoundSearchDirection>0)
			{
				if(!IsamNextKey(File,Key,UserDatRef,UserKey))
					return 0;
			}
			else
			if(NotFoundSearchDirection<0)
			{
				IsamPagePtr IPgPtr=NULL;
				IndexPertains * LInx=&File->IndexArr[Key];
				if(LInx->PathInd==0)
				{
					File->MyFiles->LastErrorCode=err_NoSmallerKeyFound;
					return 0;
				}
				if(!IsamGetPage(File,LInx->Path[LInx->PathInd-1].PageRef,Key,&IPgPtr))
					return 0;
				*UserDatRef=IPgPtr->ItemArray[LInx->Path[LInx->PathInd-1].ItemArrInd-1].DataRef;
				CopyKeyStr(UserKey,&IPgPtr->ItemArray[LInx->Path[LInx->PathInd-1].ItemArrInd-1].KeyStr);
			}
			else
			{
				File->MyFiles->LastErrorCode=err_NoMatching;
				return 0;
			}
		}
		else
		{
			return 0;
		}
	}
	return 1;
}

int IsamPrevKey(P_local_file File,unsigned short Key,int * UserDatRef,IsamKeyStr * UserKey)
{
	KeyNr=Key;
	if((Key>0)&&(Key<=_Files->Files[File->Index].datInfo.NrOfKeys))
	{
		int RefNr;
		IsamPagePtr IPgPtr=NULL;
		IndexPertains * LInx=&File->IndexArr[Key];
		IsamDatIndDescr * GInx=&_Files->Files[File->Index].InxDescr[Key];

		if(!IsValidIndex(File,Key))
		{
//			if(GInx->SearchForSequentialEnabled)
			{
                RecallLastKey(LInx,UserDatRef,UserKey);
				if(!IsamFindKeyAndRef(File,Key,UserDatRef,UserKey,0))
				{
					if(File->MyFiles->LastErrorCode==err_NoMatching)
					{
						return IsamFindKeyAndRef(File,Key,UserDatRef,UserKey,-1);
					}
					return 0;
				}
				ValidateIndex(File,Key);
			}
//			else
//			{
//				File->MyFiles->LastErrorCode=err_NotSequentialAccess_;
//				return 0;
//			}
		}
		LInx->SequentialOK=0;
		if(LInx->PathInd==0)
			RefNr=GInx->RootRef;
		else
		{
			if(!IsamGetPage(File,LInx->Path[LInx->PathInd-1].PageRef,Key,&IPgPtr))
				return 0;
			LInx->Path[LInx->PathInd-1].ItemArrInd--;
			if(LInx->Path[LInx->PathInd-1].ItemArrInd==0)
				RefNr=IPgPtr->BckwPageRef;
			else
				RefNr=IPgPtr->ItemArray[LInx->Path[LInx->PathInd-1].ItemArrInd-1].PageRef;
		}
		while(RefNr!=0)
		{
			if(!IsamGetPage(File,RefNr,Key,&IPgPtr))
				return 0;
			LInx->PathInd++;
			LInx->Path[LInx->PathInd-1].PageRef=RefNr;
			LInx->Path[LInx->PathInd-1].ItemArrInd=IPgPtr->ItemsOnPage;
			RefNr=IPgPtr->ItemArray[IPgPtr->ItemsOnPage-1].PageRef;
		}
		if(LInx->PathInd)
		{
			while((LInx->PathInd>1)&&(LInx->Path[LInx->PathInd-1].ItemArrInd==0))
			{
				LInx->PathInd--;
				if(!IsamGetPage(File,LInx->Path[LInx->PathInd-1].PageRef,Key,&IPgPtr))
					return 0;
			}
			if(LInx->Path[LInx->PathInd-1].ItemArrInd>0)
			{
				CopyKeyStr(UserKey,&IPgPtr->ItemArray[LInx->Path[LInx->PathInd-1].ItemArrInd-1].KeyStr);
				*UserDatRef=IPgPtr->ItemArray[LInx->Path[LInx->PathInd-1].ItemArrInd-1].DataRef;
			}
			else
			{
				LInx->PathInd=0;
			}
		}
		LInx->SequentialOK=1;
		if(!LInx->PathInd)
		{
			File->MyFiles->LastErrorCode=err_NoSmallerKeyFound;
			return 0;
		}
		SetLastKey(LInx,UserDatRef,UserKey);
	}
	else
	{
		File->MyFiles->LastErrorCode=err_BadKey;
		return 0;
	}
	return 1;
}

int IsamSearchKey(P_local_file File,unsigned short Key,int * UserDatRef,IsamKeyStr * UserKey)
{
	File->MyFiles->LastErrorCode=0;
	*UserDatRef=0;
	if(!_IsamFindKey(File,Key,UserDatRef,UserKey))
	{
		if(File->MyFiles->LastErrorCode==0)
		{
            if(File->IndexArr[Key].SequentialOK)
                ValidateIndex(File,Key);
			if(IsamNextKey(File,Key,UserDatRef,UserKey))
			{
				return 1;
			}
		}
		if((File->MyFiles->LastErrorCode==0)||((int)(File->MyFiles->LastErrorCode/100)==102))
		{
			File->MyFiles->LastErrorCode=err_KeyOrLargeNotFound;
		}
		return 0;
	}
	return 1;
}

int IsamSearchKeyAndRef(P_local_file File,unsigned short Key,int * UserDatRef,IsamKeyStr * UserKey)
{
	if(!IsamFindKeyAndRef(File,Key,UserDatRef,UserKey,1))
	{
		if(File->MyFiles->LastErrorCode==err_NoLargerKeyFound_)
		{
			File->MyFiles->LastErrorCode=0;
            if(File->IndexArr[Key].SequentialOK)
                ValidateIndex(File,Key);
			if(IsamPrevKey(File,Key,UserDatRef,UserKey))
			{
				return 1;
			}
		}
		return 0;
	}
	return 1;
}

int IsamGetBlock(P_local_files Local,int Handle,int Offset,unsigned short int Length,void * Dest)
{
	if(lseek(Handle,Offset,SEEK_SET)==Offset)
	{
		if(read(Handle,(void *)Dest,Length)==Length)
		{
			return 1;
		}
		Local->LastErrorCode=errno;
		if(!Local->LastErrorCode)
			Local->LastErrorCode=err_ReadError;
	}
	else
	{
		Local->LastErrorCode=errno;
		if(!Local->LastErrorCode)
			Local->LastErrorCode=err_SeekError;
	}
	return 0;
}

int IsamNewRec(P_local_file File,int * RefNr,unsigned short Ind)
{
	P_pds_file GF=&_Files->Files[File->Index];
    IsamDatIndDescrPtr DI=&GF->InxDescr[Ind];
	int NextFree;
	if(DI->FirstFree!=-1)
	{
		if((DI->FirstFree<=0)||(DI->FirstFree>DI->NumRec))
		{
			File->MyFiles->LastErrorCode=err_FreeListCorrupt;
			return 0;
		}
		*RefNr=DI->FirstFree;
		if(!Ind)
		{
			if(!IsamGetBlock(File->MyFiles,GF->Handle,*RefNr*DI->LenRec,sizeof(NextFree),&NextFree))
				return 0;
		}
		else
		{
			if(!IsamGetBlock(File->MyFiles,GF->InxHandle,*RefNr*GF->BlockLen+DI->BlockOfs,sizeof(NextFree),&NextFree))
				return 0;
		}
		if((NextFree==0)||(NextFree>DI->NumRec)||(NextFree<-1))
		{
			File->MyFiles->LastErrorCode=err_FreeListCorrupt;
			return 0;
		}
		DI->FirstFree=NextFree;
		DI->NumberFree--;
	}
	else
	{
		DI->NumRec++;
		*RefNr=DI->NumRec;
		if((Ind)&&(DI->NumRec>GF->MaxPages))
		{
			if(!IsamPutDummyBlock(File->MyFiles,GF->InxHandle,(GF->MaxPages+1)*GF->BlockLen,GF->BlockLen))
			{
				DI->NumRec--;
				return 0;
			}
			GF->MaxPages++;
		}
	}
	if(!IsamPutInfoRec(File->MyFiles,GF,Ind,0))
		return 0;
	return 1;
}

int IsamAddRec(P_local_file File,int * RefNr,void * Source)
{
	P_pds_file GF=&_Files->Files[File->Index];
    IsamDatIndDescrPtr DI=&GF->InxDescr[0];
	int OldFirstFree=DI->FirstFree;
	if(!IsamNewRec(File,RefNr,0))
		return 0;
	if(!IsamPutBlock(File->MyFiles,GF->Handle,*RefNr*DI->LenRec,DI->LenRec,Source))
	{
		DI->FirstFree=OldFirstFree;
		DI->NumberFree++;
		IsamPutInfoRec(File->MyFiles,GF,0,0);
		return 0;
	}
	return 1;
};

int IsamPutRec(P_local_file File,int RefNr,void * Source)
{
	P_pds_file GF=&_Files->Files[File->Index];
    IsamDatIndDescrPtr DI=&GF->InxDescr[0];
    if((RefNr<=0)||(RefNr>DI->NumRec))
    {
        File->MyFiles->LastErrorCode=err_BadRecNum;
        return 0;
    }
	if(!IsamPutBlock(File->MyFiles,GF->Handle,RefNr*DI->LenRec,DI->LenRec,Source))
	{
	    return 0;
	}
	return 1;
}

int IsamUpdatePage(P_local_file File,IsamPagePtr PgPtr)
{
	P_page_buffer PageBuffer=GetPageBufferFromIsamPagePtr(PgPtr);
	return IsamWritePage(File,PageBuffer);
}

int IsamNewPage(P_local_file File,int * PRef,IsamPagePtr * IPgPtr,unsigned short Key)
{
	if(!IsamNewRec(File,PRef,Key))
		return 0;
	P_page_buffer PageBuffer=GetPageBuffer(File,Key,*PRef);
	if(PageBuffer==NULL)
		return 0;
	PageBuffer->Valid=1;
	*IPgPtr=GetIsamPagePtrFromBuffer(PageBuffer);
	return 1;
}

typedef struct IAKParam IAKParam,*P_IAKParam;

struct IAKParam {
	int					IPgRef2;
	IsamItem			IItem1;
	IsamItem			IItem2;
	IsamKeyStr *		UserKey;
	int *				UserDatRef;
	P_local_file		File;
	IsamPagePtr			IPgPtr1;
	IsamPagePtr			IPgPtr2;
	IsamDatIndDescrPtr 	DI;
	short				C;
	short				K;
	short				L;
	unsigned short		Key;
	uint8_t 			OverRun;
	uint8_t				ADK;
	uint8_t				OK;
	uint8_t				IsOK;
	uint8_t 			right;
};

void IAKBalance(P_IAKParam Prm,short int * RV,/*int * IPgRef1V,*/short * R,uint8_t * right,IsamPagePtr * IPgPtr1V,IsamPagePtr * IPgPtr3)
{
	int TRef;
	short I,K,M,N;
	unsigned short CurPageSize=Prm->DI->UsedPageSize;
	if(*right)
	{
		(*RV)++;
		(*R)++;
		TRef=(*IPgPtr1V)->ItemArray[*RV-1].PageRef;
		(*IPgPtr1V)->ItemArray[*RV-1].PageRef=(*IPgPtr3)->BckwPageRef;
		K=2+CurPageSize+(*IPgPtr3)->ItemsOnPage;
		M=(K>>1)+1;
		K=K-M-(*IPgPtr3)->ItemsOnPage;
		for(I=(*IPgPtr3)->ItemsOnPage-1;I>=0;I--)
		{
			(*IPgPtr3)->ItemArray[I+K]=(*IPgPtr3)->ItemArray[I]; // se snizenim ukazatele pole o 1 uz pocitano v promennych cyklu
		}
		(*IPgPtr3)->ItemArray[K-1]=(* IPgPtr1V)->ItemArray[*RV-1];
		if(M==*R)
		{
			for(I=*R;I<=CurPageSize;I++)
			{
				(*IPgPtr3)->ItemArray[I-*R]=Prm->IPgPtr1->ItemArray[I-1];
			}
			(*IPgPtr1V)->ItemArray[*RV-1]=Prm->IItem1;
		}
		else
		if(M>*R)
		{
			for(I=M;I<=CurPageSize;I++)
			{
				(*IPgPtr3)->ItemArray[I-M]=Prm->IPgPtr1->ItemArray[I-1];
			}
			(*IPgPtr1V)->ItemArray[*RV-1]=Prm->IPgPtr1->ItemArray[M-2];
			for(I=M-2;I>=*R;I--)
			{
				Prm->IPgPtr1->ItemArray[I]=Prm->IPgPtr1->ItemArray[I-1];
			}
			Prm->IPgPtr1->ItemArray[*R-1]=Prm->IItem1;
		}
		else
		{
			for(I=*R;I<=CurPageSize;I++)
			{
				(*IPgPtr3)->ItemArray[I-M]=Prm->IPgPtr1->ItemArray[I-1];
			}
			(*IPgPtr3)->ItemArray[*R-M-1]=Prm->IItem1;
			for(I=M+1;I<*R;I++)
			{
				(*IPgPtr3)->ItemArray[I-M-1]=Prm->IPgPtr1->ItemArray[I-1];
			}
			(*IPgPtr1V)->ItemArray[*RV-1]=Prm->IPgPtr1->ItemArray[M-1];
		}
		(*IPgPtr3)->BckwPageRef=(*IPgPtr1V)->ItemArray[*RV-1].PageRef;
		(*IPgPtr1V)->ItemArray[*RV-1].PageRef=TRef;
		Prm->IPgPtr1->ItemsOnPage=M-1;
		(*IPgPtr3)->ItemsOnPage+=K;
	}
	else
	{
		(*R)++;
		TRef=(*IPgPtr1V)->ItemArray[*RV-1].PageRef;
		(*IPgPtr1V)->ItemArray[*RV-1].PageRef=Prm->IPgPtr1->BckwPageRef;
		N=(*IPgPtr3)->ItemsOnPage+1;
		M=(2+CurPageSize+N)>>1;
		K=M-N;
		(*IPgPtr3)->ItemArray[N-1]=(*IPgPtr1V)->ItemArray[*RV-1];
		if(K==*R)
		{
			for(I=0;I<K-1;I++)
			{
				(*IPgPtr3)->ItemArray[I+N]=Prm->IPgPtr1->ItemArray[I]; // se snizenim ukazatele pole o 1 uz pocitano v promennych cyklu
			}
			for(I=*R;I<=CurPageSize;I++)
			{
				Prm->IPgPtr1->ItemArray[I-*R]=Prm->IPgPtr1->ItemArray[I-1];
			}
			(*IPgPtr1V)->ItemArray[*RV-1]=Prm->IItem1;
		}
		else
		if(K<*R)
		{
			for(I=0;I<K-1;I++)
			{
				(*IPgPtr3)->ItemArray[I+N]=Prm->IPgPtr1->ItemArray[I]; // se snizenim ukazatele pole o 1 uz pocitano v promennych cyklu
			}
			(*IPgPtr1V)->ItemArray[*RV-1]=Prm->IPgPtr1->ItemArray[K-1];
			for(I=K;I<*R-1;I++)
			{
				Prm->IPgPtr1->ItemArray[I-K]=Prm->IPgPtr1->ItemArray[I]; // se snizenim ukazatele pole o 1 uz pocitano v promennych cyklu
			}
			Prm->IPgPtr1->ItemArray[*R-K-1]=Prm->IItem1;
			for(I=*R;I<=CurPageSize;I++)
			{
				Prm->IPgPtr1->ItemArray[I-K]=Prm->IPgPtr1->ItemArray[I-1];
			}
		}
		else
		{
			for(I=0;I<*R-1;I++)
			{
				(*IPgPtr3)->ItemArray[I+N]=Prm->IPgPtr1->ItemArray[I]; // se snizenim ukazatele pole o 1 uz pocitano v promennych cyklu
			}
			(*IPgPtr3)->ItemArray[*R+N-1]=Prm->IItem1;
			for(I=*R;I<K-1;I++)
			{
				(*IPgPtr3)->ItemArray[I+N]=Prm->IPgPtr1->ItemArray[I-1];
			}
			(*IPgPtr1V)->ItemArray[*RV-1]=Prm->IPgPtr1->ItemArray[K-2];
			for(I=K;I<=CurPageSize;I++)
			{
				Prm->IPgPtr1->ItemArray[I-K]=Prm->IPgPtr1->ItemArray[I-1];
			}
		}
		Prm->IPgPtr1->BckwPageRef=(*IPgPtr1V)->ItemArray[*RV-1].PageRef;
		(*IPgPtr1V)->ItemArray[*RV-1].PageRef=TRef;
		Prm->IPgPtr1->ItemsOnPage=CurPageSize-(K-1);
		(*IPgPtr3)->ItemsOnPage+=K;
	}
}

int IAKBalanceDone(P_IAKParam Prm,short int * RV,int * IPgRef1V, short * R)
{
	int IPgRef3;
	IsamPagePtr IPgPtr1V;
	IsamPagePtr IPgPtr3;
	uint8_t right;
	Prm->IsOK=1;
	if(*RV==-1)
	{
		return 0;
	}
	if(!IsamGetPage(Prm->File,*IPgRef1V,Prm->Key,&IPgPtr1V))
	{
		Prm->IsOK=0;
		return 0;
	}
	right=*RV<IPgPtr1V->ItemsOnPage;
	if(right)
	{
		IPgRef3=IPgPtr1V->ItemArray[*RV].PageRef;
	}
	else
	{
		if(*RV>1)
		{
			IPgRef3=IPgPtr1V->ItemArray[*RV-2].PageRef;
		}
		else
		{
			IPgRef3=IPgPtr1V->BckwPageRef;
		}
	}
	if(!IsamGetPage(Prm->File,IPgRef3,Prm->Key,&IPgPtr3))
	{
		Prm->IsOK=0;
		return 0;
	}
	if(IPgPtr3->ItemsOnPage==Prm->DI->UsedPageSize)
		return 0;
	IAKBalance(Prm,RV,/*IPgRef1V,*/R,&right,&IPgPtr1V,&IPgPtr3);
	int Result=1;
	if(!IsamUpdatePage(Prm->File,IPgPtr3))
		Result=0;
	if(!IsamUpdatePage(Prm->File,IPgPtr1V))
		Result=0;
	Prm->OverRun=0;
	return Result;
}

int IAKInsert(P_IAKParam Prm,short int * RV,int * IPgRef1V, short * R,int * IPgRef1)
{
	short I;
	int Result=1;
	unsigned short CurPageSize=Prm->DI->UsedPageSize;
	if(!IsamGetPage(Prm->File,*IPgRef1,Prm->Key,&Prm->IPgPtr1))
		return 0;
	if(Prm->IPgPtr1->ItemsOnPage<CurPageSize)
	{
		Prm->IPgPtr1->ItemsOnPage++;
		memmove(&Prm->IPgPtr1->ItemArray[*R+1],&Prm->IPgPtr1->ItemArray[*R],(Prm->IPgPtr1->ItemsOnPage-(*R+1))*sizeof(IsamItem));
		Prm->IPgPtr1->ItemArray[*R]=Prm->IItem1;
		Prm->OverRun=0;
	}
	else
	{
		if(!IAKBalanceDone(Prm,RV,IPgRef1V,R))
		{
			if(!Prm->IsOK)
				return 0;
			if(!IsamNewPage(Prm->File,&Prm->IPgRef2,&Prm->IPgPtr2,Prm->Key))
				return 0;
			if(*R<=CurPageSize>>1)
			{
				if(*R==CurPageSize>>1)
				{
					Prm->IItem2=Prm->IItem1;
				}
				else
				{
					Prm->IItem2=Prm->IPgPtr1->ItemArray[(CurPageSize>>1)-1];
					for(I=(CurPageSize>>1)-1;I>*R;I--)
					{
						Prm->IPgPtr1->ItemArray[I]=Prm->IPgPtr1->ItemArray[I-1];
					}
					Prm->IPgPtr1->ItemArray[*R]=Prm->IItem1;
				}
				for(I=0;I<CurPageSize>>1;I++)
				{
					Prm->IPgPtr2->ItemArray[I]=Prm->IPgPtr1->ItemArray[I+(CurPageSize>>1)];
				}
			}
			else
			{
				*R=*R-(CurPageSize>>1);
				Prm->IItem2=Prm->IPgPtr1->ItemArray[CurPageSize>>1];
				for(I=0;I<*R-1;I++)
				{
					Prm->IPgPtr2->ItemArray[I]=Prm->IPgPtr1->ItemArray[I+(CurPageSize>>1)+1];
				}
				Prm->IPgPtr2->ItemArray[*R-1]=Prm->IItem1;
				for(I=*R;I<CurPageSize>>1;I++)
				{
					Prm->IPgPtr2->ItemArray[I]=Prm->IPgPtr1->ItemArray[I+(CurPageSize>>1)];
				}
			}
			Prm->IPgPtr1->ItemsOnPage=CurPageSize>>1;
			Prm->IPgPtr2->ItemsOnPage=CurPageSize>>1;
			Prm->IPgPtr2->BckwPageRef=Prm->IItem2.PageRef;
			Prm->IItem2.PageRef=Prm->IPgRef2;
			Prm->IItem1=Prm->IItem2;
			if(!IsamUpdatePage(Prm->File,Prm->IPgPtr2))
				Result=0;
		}
	}
	if(!IsamUpdatePage(Prm->File,Prm->IPgPtr1))
		Result=0;
	return Result;
}

int SearchGen=0;

int IAKSearch(int IPgRef1,short int RV,int IPgRef1V,P_IAKParam Prm)
{
	SearchGen++;
	short R;
	uint8_t IsXL;
	if(IPgRef1==0)
	{
		Prm->OverRun=1;
		CopyKeyStr(&Prm->IItem1.KeyStr,Prm->UserKey);
		Prm->IItem1.DataRef=*Prm->UserDatRef;
		Prm->IItem1.PageRef=0;
	}
	else
	{
		if(!IsamGetPage(Prm->File,IPgRef1,Prm->Key,&Prm->IPgPtr1))
			return 0;
		Prm->L=1;
		R=Prm->IPgPtr1->ItemsOnPage;
		Prm->ADK=Prm->DI->AllowDupKeys;
		IsXL=Prm->DI->XlatKeys;
		do {
			Prm->K=(Prm->L+R) >> 1;
			Prm->C=IsamCompKeys(*(Prm->UserKey),Prm->IPgPtr1->ItemArray[Prm->K-1].KeyStr,*Prm->UserDatRef,Prm->IPgPtr1->ItemArray[Prm->K-1].DataRef,Prm->ADK,IsXL);
			if(Prm->C<=0)
				R=Prm->K-1;
			if(Prm->C>=0)
				Prm->L=Prm->K+1;
		} while (R>=Prm->L);
		if(Prm->L-R>1)
		{
			Prm->OK=0;
			Prm->OverRun=0;
		}
		else
		{
			if(R==0)
			{
				if(!IAKSearch(Prm->IPgPtr1->BckwPageRef,R,IPgRef1,Prm))
					return 0;
			}
			else
			{
				if(!IAKSearch(Prm->IPgPtr1->ItemArray[R-1].PageRef,R,IPgRef1,Prm))
					return 0;
			}
			if(Prm->OverRun)
			{
				if(!IAKInsert(Prm,&RV,&IPgRef1V,&R,&IPgRef1))
					return 0;
			}
		}
	}
	SearchGen--;
	return 1;
}

int IsamAddKey(P_local_file File,unsigned short Key,int UserDatRef,IsamKeyStr * UserKey)
{
	KeyNr=Key;
	SearchGen=0;
	int	IPgRef1;
	IAKParam IAKP;
	memset(&IAKP,0,sizeof(IAKP));
	IAKP.UserKey=UserKey;
	IAKP.UserDatRef=&UserDatRef;
	IAKP.Key=Key;
	IAKP.File=File;

	P_pds_file GF=&_Files->Files[File->Index];
	if((Key>0)&&(Key<=GF->datInfo.NrOfKeys))
	{
    	IAKP.DI=&GF->InxDescr[Key];
		IndexPertains * LInx=&File->IndexArr[Key];
		if(UserKey->Length>IAKP.DI->KeyLen)
			UserKey->Length=IAKP.DI->KeyLen;
		LInx->SequentialOK=0;
		IAKP.OK=1;
		if(!IAKSearch(IAKP.DI->RootRef,-1,0,&IAKP))
			return 0;
		if(IAKP.OverRun)
		{
			IPgRef1=IAKP.DI->RootRef;
			if(!IsamNewPage(File,&IAKP.DI->RootRef,&IAKP.IPgPtr1,Key))
				return 0;
			IAKP.IPgPtr1->ItemsOnPage=1;
			IAKP.IPgPtr1->BckwPageRef=IPgRef1;
			IAKP.IPgPtr1->ItemArray[0]=IAKP.IItem1;
			IsamUpdatePage(File,IAKP.IPgPtr1);
		}
		if(!IAKP.OK)
		{
		    const char * FuncName="nezname";
		    if(File->MyFiles)
		    {
		        if(File->MyFiles->FuncName)
                    FuncName=File->MyFiles->FuncName;
		    }
			message(1,"%d Pokus o pridani duplicitniho klice v indexu %d veta %d pro soubor %s (v %s)\n",getpid(),Key,UserDatRef,GF->Name,FuncName);
			return 0;
		}
		IAKP.DI->NumKeys++;
		IsamPutInfoRec(File->MyFiles,&_Files->Files[File->Index],Key,0);
	}
	else
	{
		File->MyFiles->LastErrorCode=err_BadKey;
		return 0;
	}
	return 1;
}

int IsamGetApprRelPos(P_local_file File,unsigned short Key,unsigned short * RelPos,unsigned short Scale,IsamKeyStr * UserKey,int UserDatRef)
{
	int TempL,RefNr;
	short C,K,L,R;
	IsamPagePtr IPgPtr;
	unsigned short EstEltNr,EstNrOfElts,RootItemNr,ItemNrOnSecPage,NrOfItemsOnRoot,NrOfItemsOnSecPage,AvgPageSizeSecLevel;
	uint8_t Stop;
	if(!Scale)
	{
		File->MyFiles->LastErrorCode=err_InvalidScale;
		return 0;
	}
	P_pds_file GF=&_Files->Files[File->Index];
    IsamDatIndDescrPtr DI=&GF->InxDescr[Key];
	if(DI->RootRef==0)
	{
		File->MyFiles->LastErrorCode=err_IndexEmpty;
		return 0;
	}
	AvgPageSizeSecLevel=DI->UsedPageSize-(DI->UsedPageSize>>2);
	NrOfItemsOnSecPage=AvgPageSizeSecLevel;
	RootItemNr=0xFFFF;
	ItemNrOnSecPage=0;
	Stop=0;
	RefNr=DI->RootRef;
	NrOfItemsOnRoot=0;
	while((RefNr)&&(!Stop))
	{
		if(!IsamGetPage(File,RefNr,Key,&IPgPtr))
			return 0;
		L=1;
		R=IPgPtr->ItemsOnPage;
		do {
			K=(L+R)>>1;
			C=IsamCompKeys(*UserKey,IPgPtr->ItemArray[K-1].KeyStr,UserDatRef,IPgPtr->ItemArray[K-1].DataRef,DI->AllowDupKeys,DI->XlatKeys);
			if(C<=0)
				R=K-1;
			if(C>=0)
				L=K+1;
		} while (R>=L);
		if(L-R>1)
			Stop=1;
		else
		if(R==0)
			RefNr=IPgPtr->BckwPageRef;
		else
			RefNr=IPgPtr->ItemArray[R-1].PageRef;
		if(RootItemNr==0xFFFF)
		{
			NrOfItemsOnRoot=IPgPtr->ItemsOnPage;
			if(L-R>1)
				RootItemNr=K;
			else
				RootItemNr=R;
		}
		else
		{
			NrOfItemsOnSecPage=IPgPtr->ItemsOnPage;
			if(L-R>1)
				ItemNrOnSecPage=K;
			else
				ItemNrOnSecPage=R;
			Stop=1;
		}
	}
	EstNrOfElts=(AvgPageSizeSecLevel+1)*(NrOfItemsOnRoot+1);
	if((ItemNrOnSecPage==1)&&(RootItemNr==0))
	{
		EstEltNr=1;
	}
	else
	{
		TempL=(int)((int)(ItemNrOnSecPage)*(int)(((int)AvgPageSizeSecLevel+1)<<1)/(int)NrOfItemsOnSecPage);
		EstEltNr=RootItemNr*(AvgPageSizeSecLevel+1)+(TempL-(TempL>>1));
	}
	*RelPos=(unsigned short)((int)EstEltNr*((int)Scale+1)/(int)EstNrOfElts);
	if(*RelPos>Scale)
		*RelPos=Scale;
	return 1;
}

int IsamGetApprKeyAndRef(P_local_file File,unsigned short Key,unsigned short RelPos,unsigned short Scale,IsamKeyStr * UserKey,int * UserDatRef)
{
	unsigned short EstNrOfElts,EstEltNr,EstRootItemNr,EstItemNrOnSecPage,CorItemNrOnSecPage,AvgPageSizeSecLevel;
	IsamPagePtr IPgPtr;
	int TempL;
	if((Scale==0)||(RelPos>Scale))
	{
		File->MyFiles->LastErrorCode=err_InvalidPosition;
		return 0;
	}
	P_pds_file GF=&_Files->Files[File->Index];
    IsamDatIndDescrPtr DI=&GF->InxDescr[Key];
	if(DI->RootRef==0)
	{
		File->MyFiles->LastErrorCode=err_IndexEmpty_;
		return 0;
	}
	if(!IsamGetPage(File,DI->RootRef,Key,&IPgPtr))
	{
		return 0;
	}
	AvgPageSizeSecLevel=DI->UsedPageSize-(DI->UsedPageSize>>2);
	if(IPgPtr->BckwPageRef==0)
		AvgPageSizeSecLevel=0;
	EstNrOfElts=(AvgPageSizeSecLevel+1)*(IPgPtr->ItemsOnPage+1);
	EstEltNr=((int)EstNrOfElts*(int)RelPos)/((int)Scale+1)+EstNrOfElts/Scale-((EstNrOfElts/Scale)>>1);
	EstRootItemNr=EstEltNr/(AvgPageSizeSecLevel+1);
	EstItemNrOnSecPage=EstEltNr%(AvgPageSizeSecLevel+1);
	if((IPgPtr->BckwPageRef==0)||(EstItemNrOnSecPage==0))
	{
		if(EstRootItemNr==0)
		{
			(*UserKey).Length=0;
			*UserDatRef=0;
		}
		else
		{
			CopyKeyStr(UserKey,&IPgPtr->ItemArray[EstRootItemNr-1].KeyStr);
			*UserDatRef=IPgPtr->ItemArray[EstRootItemNr-1].DataRef;
		}
		return 1;
	}
	if(EstRootItemNr==0)
	{
		if(!IsamGetPage(File,IPgPtr->BckwPageRef,Key,&IPgPtr))
			return 0;
	}
	else
	{
		if(!IsamGetPage(File,IPgPtr->ItemArray[EstRootItemNr-1].PageRef,Key,&IPgPtr))
			return 0;
	}
	if((EstItemNrOnSecPage==2)&&(EstRootItemNr==0))
	{
		CorItemNrOnSecPage=2;
	}
	else
	{
		TempL=(((int)EstItemNrOnSecPage*(int)IPgPtr->ItemsOnPage)<<1)/(int)AvgPageSizeSecLevel;
		CorItemNrOnSecPage=TempL-(TempL>>1);
	}
	if(CorItemNrOnSecPage==0)
		CorItemNrOnSecPage++;
	CopyKeyStr(UserKey,&IPgPtr->ItemArray[CorItemNrOnSecPage-1].KeyStr);
	*UserDatRef=IPgPtr->ItemArray[CorItemNrOnSecPage-1].DataRef;
	return 1;
}

int IsamDeleteRecOrPage(P_local_file File,int RefNr,unsigned short Ind)
{
	P_pds_file GF=&_Files->Files[File->Index];
	if(File->Index==-1)
	{
		File->MyFiles->LastErrorCode=err_NotOpen;
		return 0;
	}
	if(Ind>GF->datInfo.NrOfKeys)
	{
		File->MyFiles->LastErrorCode=err_BadKey;
		return 0;
	}
	IsamDatIndDescrPtr DI=&GF->InxDescr[Ind];
	if(Ind==0)
	{
		int DeleteArr[2]={DI->FirstFree,-1};
		if(!IsamPutBlock(File->MyFiles,GF->Handle,RefNr*DI->LenRec,sizeof(DeleteArr),&DeleteArr))
			return 0;
	}
	else
	{
		if(!IsamPutBlock(File->MyFiles,GF->InxHandle,RefNr*GF->BlockLen+DI->BlockOfs,sizeof(int),&DI->FirstFree))
			return 0;
	}
	DI->FirstFree=RefNr;
	DI->NumberFree++;
	if(!IsamPutInfoRec(File->MyFiles,GF,Ind,0))
		return 0;
	return 1;
}

int IsamDeleteRec(P_local_file File,int RefNr)
{
    return IsamDeleteRecOrPage(File,RefNr,0);
}

int IsamReturnPage(P_local_file File,IsamPagePtr PgPtr)
{
	P_page_buffer PageBuffer=GetPageBufferFromIsamPagePtr(PgPtr);
	if(!IsamDeleteRecOrPage(File,PageBuffer->PRef,PageBuffer->Key))
		return 0;
	return 1;
}

typedef struct IDKParam IDKParam,*P_IDKParam;

struct IDKParam {
	IsamKeyStr *		UserKey;
	int *				UserDatRef;
	P_local_file		File;
	IsamDatIndDescrPtr 	DI;
	unsigned short		Key;
	uint8_t 			UnderRun;
	uint8_t				OK;
};

int IDKUnderFlow(int RefNr,int IPgRef2,short int R,P_IDKParam Prm)
{
	short int I,K,IItem0;
	int IPgRef0;
	IsamPagePtr IPgPtr,IPgPtr2,IPgPtr0;
	unsigned short int CurPageSize=Prm->DI->UsedPageSize;
	if(!IsamGetPage(Prm->File,RefNr,Prm->Key,&IPgPtr))
		return 0;
	if(!IsamGetPage(Prm->File,IPgRef2,Prm->Key,&IPgPtr2))
		return 0;
	if(R<IPgPtr->ItemsOnPage)
	{
		R++;
		IPgRef0=IPgPtr->ItemArray[R-1].PageRef;
		if(!IsamGetPage(Prm->File,IPgRef0,Prm->Key,&IPgPtr0))
			return 0;
		K=(IPgPtr0->ItemsOnPage-(CurPageSize>>1)+1)>>1;
		IPgPtr2->ItemArray[(CurPageSize>>1)-1]=IPgPtr->ItemArray[R-1];
		IPgPtr2->ItemArray[(CurPageSize>>1)-1].PageRef=IPgPtr0->BckwPageRef;
		if(K>0)
		{
			for(I=0;I<K-1;I++)
			{
				IPgPtr2->ItemArray[I+(CurPageSize>>1)]=IPgPtr0->ItemArray[I];
			}
			IPgPtr->ItemArray[R-1]=IPgPtr0->ItemArray[K-1];
			IPgPtr->ItemArray[R-1].PageRef=IPgRef0;
			IPgPtr0->BckwPageRef=IPgPtr0->ItemArray[K-1].PageRef;
			IPgPtr0->ItemsOnPage=IPgPtr0->ItemsOnPage-K;
			for(I=0;I<IPgPtr0->ItemsOnPage;I++)
			{
				IPgPtr0->ItemArray[I]=IPgPtr0->ItemArray[I+K];
			}
			IPgPtr2->ItemsOnPage=(CurPageSize>>1)+K-1;
			Prm->UnderRun=0;
			IsamUpdatePage(Prm->File,IPgPtr0);
		}
		else
		{
			for(I=0;I<CurPageSize>>1;I++)
			{
				IPgPtr2->ItemArray[I+(CurPageSize>>1)]=IPgPtr0->ItemArray[I];
			}
			for(I=R-1;I<IPgPtr->ItemsOnPage;I++)
			{
				IPgPtr->ItemArray[I]=IPgPtr->ItemArray[I+1];
			}
			IPgPtr2->ItemsOnPage=CurPageSize;
			IPgPtr->ItemsOnPage--;
			if(!IsamReturnPage(Prm->File,IPgPtr0))
				return 0;
			if(IPgPtr->ItemsOnPage<CurPageSize>>1)
				Prm->UnderRun=1;
			else
				Prm->UnderRun=0;
		}
		IsamUpdatePage(Prm->File,IPgPtr2);
	}
	else
	{
		if(R==1)
			IPgRef0=IPgPtr->BckwPageRef;
		else
			IPgRef0=IPgPtr->ItemArray[R-2].PageRef;
		if(!IsamGetPage(Prm->File,IPgRef0,Prm->Key,&IPgPtr0))
			return 0;
		IItem0=IPgPtr0->ItemsOnPage+1;

		K=(IItem0-(CurPageSize>>1))>>1;
		if(K>0)
		{
			for(I=(CurPageSize>>1)-2;I>=0;I--)
			{
				IPgPtr2->ItemArray[I+K]=IPgPtr2->ItemArray[I];
			}
			IPgPtr2->ItemArray[K-1]=IPgPtr->ItemArray[R-1];
			IPgPtr2->ItemArray[K-1].PageRef=IPgPtr2->BckwPageRef;
			IItem0-=K;
			for(I=K-2;I>=0;I--)
			{
				IPgPtr2->ItemArray[I]=IPgPtr0->ItemArray[I+IItem0];
			}
			IPgPtr2->BckwPageRef=IPgPtr0->ItemArray[IItem0-1].PageRef;
			IPgPtr->ItemArray[R-1]=IPgPtr0->ItemArray[IItem0-1];
			IPgPtr->ItemArray[R-1].PageRef=IPgRef2;
			IPgPtr0->ItemsOnPage=IItem0-1;
			IPgPtr2->ItemsOnPage=((CurPageSize>>1)+K)-1;
			Prm->UnderRun=0;
			IsamUpdatePage(Prm->File,IPgPtr2);
		}
		else
		{
			IPgPtr0->ItemArray[IItem0-1]=IPgPtr->ItemArray[R-1];
			IPgPtr0->ItemArray[IItem0-1].PageRef=IPgPtr2->BckwPageRef;
			for(I=0;I<(CurPageSize>>1)-1;I++)
			{
				IPgPtr0->ItemArray[I+IItem0]=IPgPtr2->ItemArray[I];
			}
			IPgPtr0->ItemsOnPage=CurPageSize;
			IPgPtr->ItemsOnPage--;
			if(! IsamReturnPage(Prm->File,IPgPtr2))
				return 0;
			if(IPgPtr->ItemsOnPage<CurPageSize>>1)
				Prm->UnderRun=1;
			else
				Prm->UnderRun=0;
		}
		IsamUpdatePage(Prm->File,IPgPtr0);
	}
	IsamUpdatePage(Prm->File,IPgPtr);
	return 1;
}

int IDKDelete1(int IPgRef2,int * RefNr,short int * K,IsamPagePtr * IPgPtr,P_IDKParam Prm)
{
	short int C;
	int IPgRef00;
	IsamPagePtr IPgPtr2;
	if(!IsamGetPage(Prm->File,IPgRef2,Prm->Key,&IPgPtr2))
		return 0;
	IPgRef00=IPgPtr2->ItemArray[IPgPtr2->ItemsOnPage-1].PageRef;
	if(IPgRef00!=0)
	{
		C=IPgPtr2->ItemsOnPage;
		if(!IDKDelete1(IPgRef00,RefNr,K,IPgPtr,Prm))
			return 0;
		if(Prm->UnderRun)
			IDKUnderFlow(IPgRef2,IPgRef00,C,Prm);
	}
	else
	{
		if(!IsamGetPage(Prm->File,*RefNr,Prm->Key,IPgPtr))
			return 0;
		IPgPtr2->ItemArray[IPgPtr2->ItemsOnPage-1].PageRef=(*IPgPtr)->ItemArray[*K-1].PageRef;
		(*IPgPtr)->ItemArray[*K-1]=IPgPtr2->ItemArray[IPgPtr2->ItemsOnPage-1];
		IPgPtr2->ItemsOnPage--;
		if(IPgPtr2->ItemsOnPage<Prm->DI->UsedPageSize>>1)
			Prm->UnderRun=1;
		else
			Prm->UnderRun=0;
		IsamUpdatePage(Prm->File,*IPgPtr);
		IsamUpdatePage(Prm->File,IPgPtr2);
	}
	return 1;
}

int IDKDelete2(int RefNr,P_IDKParam Prm)
{
	short int C,K,L,R;
	int IPgRef00;
	IsamPagePtr IPgPtr;
	uint8_t ADK;
	uint8_t IsXL;

	if(RefNr==0)
	{
		Prm->OK=0;
		Prm->UnderRun=0;
	}
	else
	{
		if(!IsamGetPage(Prm->File,RefNr,Prm->Key,&IPgPtr))
			return 0;
		L=1;
		R=IPgPtr->ItemsOnPage;
		ADK=Prm->DI->AllowDupKeys;
		IsXL=Prm->DI->XlatKeys;

		do
		{
			K=(L+R)>>1;
			C=IsamCompKeys(*Prm->UserKey,IPgPtr->ItemArray[K-1].KeyStr,*Prm->UserDatRef,IPgPtr->ItemArray[K-1].DataRef,ADK,IsXL);
			if(C<=0)
				R=K-1;
			if(C>=0)
				L=K+1;
		} while (L<=R);
		if(R==0)
			IPgRef00=IPgPtr->BckwPageRef;
		else
			IPgRef00=IPgPtr->ItemArray[R-1].PageRef;
		if(L-R>1)
		{
			*Prm->UserDatRef=IPgPtr->ItemArray[K-1].DataRef;
			if(IPgRef00==0)
			{
				IPgPtr->ItemsOnPage--;
				if(IPgPtr->ItemsOnPage<Prm->DI->UsedPageSize>>1)
					Prm->UnderRun=1;
				else
					Prm->UnderRun=0;
				memmove(&IPgPtr->ItemArray[K-1],&IPgPtr->ItemArray[K],(IPgPtr->ItemsOnPage-K+1)*sizeof(IsamItem));
				IsamUpdatePage(Prm->File,IPgPtr);
			}
			else
			{
				if(!IDKDelete1(IPgRef00,&RefNr,&K,&IPgPtr,Prm))
					return 0;
				if(Prm->UnderRun)
					IDKUnderFlow(RefNr,IPgRef00,R,Prm);
			}
		}
		else
		{
			if(!IDKDelete2(IPgRef00,Prm))
				return 0;
			if(Prm->UnderRun)
				IDKUnderFlow(RefNr,IPgRef00,R,Prm);
		}
	}
	return 1;
}

int IsamDeleteKey(P_local_file File,unsigned short Key,int UserDatRef,IsamKeyStr * UserKey)
{
	IDKParam IDKP;
	memset(&IDKP,0,sizeof(IDKP));
	IDKP.UserKey=UserKey;
	IDKP.UserDatRef=&UserDatRef;
	IDKP.Key=Key;
	IDKP.File=File;
	P_pds_file GF=&_Files->Files[File->Index];
	IsamPagePtr IPgPtr;
	if((Key>0)&&(Key<=GF->datInfo.NrOfKeys))
	{
    	IDKP.DI=&GF->InxDescr[Key];
		IndexPertains * LInx=&File->IndexArr[Key];
		if(UserKey->Length>IDKP.DI->KeyLen)
			UserKey->Length=IDKP.DI->KeyLen;
		LInx->SequentialOK=0;
		IDKP.OK=1;
		if(!IDKDelete2(IDKP.DI->RootRef,&IDKP))
			return 0;
		if(IDKP.UnderRun)
		{
			if(!IsamGetPage(File,IDKP.DI->RootRef,Key,&IPgPtr))
				return 0;
			if(IPgPtr->ItemsOnPage==0)
			{
				IDKP.DI->RootRef=IPgPtr->BckwPageRef;
				if(! IsamReturnPage(File,IPgPtr))
					return 0;
			}
		}
		if(!IDKP.OK)
		{
			return 0;
		}
		IDKP.DI->NumKeys--;
		IsamPutInfoRec(File->MyFiles,&_Files->Files[File->Index],Key,0);
	}
	else
	{
		File->MyFiles->LastErrorCode=err_BadKey;
		return 0;
	}
	return 1;
}

typedef struct KEParam KEParam,*P_KEParam;

struct KEParam {
	IsamKeyStr *		UserKey;
	int *				UserDatRef;
	P_local_file		File;
	IsamDatIndDescrPtr 	DI;
	unsigned short		Key;
	int					C,K,L,R;
	IsamPagePtr			IPgPtr;
	uint8_t				ADK;
	uint8_t				IsXL;
	uint8_t	*			Result;
};

int KESearch(int PRef,P_KEParam Prm)
{
	if(PRef==0)
	{
		return 1;
	}
	else
	{
		if(!IsamGetPage(Prm->File,PRef,Prm->Key,&Prm->IPgPtr))
		{
			return 0;
		}
		Prm->L=1;
		Prm->R=Prm->IPgPtr->ItemsOnPage;
		Prm->ADK=Prm->DI->AllowDupKeys;
		if(Prm->UserDatRef==0)
			Prm->ADK=0;
		Prm->IsXL=Prm->DI->XlatKeys;
		do {
			Prm->K=(Prm->L+Prm->R)>>1;
			Prm->C=IsamCompKeys(*Prm->UserKey,Prm->IPgPtr->ItemArray[Prm->K-1].KeyStr,*Prm->UserDatRef,Prm->IPgPtr->ItemArray[Prm->K-1].DataRef,Prm->ADK,Prm->IsXL);
			if(Prm->C<=0)
				Prm->R=Prm->K-1;
			if(Prm->C>=0)
				Prm->L=Prm->K+1;
		} while (Prm->L<=Prm->R);
		if(Prm->L-Prm->R>1)
		{
			*(Prm->Result)=1;
		}
		else
		{
			if(Prm->R==0)
				KESearch(Prm->IPgPtr->BckwPageRef,Prm);
			else
				KESearch(Prm->IPgPtr->ItemArray[Prm->R-1].PageRef,Prm);
		}
	}
	return 1;
}

int KeyExists(P_local_file File,unsigned short Key,int UserDatRef,IsamKeyStr * UserKey,uint8_t * Result)
{
	*Result=0;
	KEParam KEP;
	memset(&KEP,0,sizeof(KEP));
	KEP.UserKey=UserKey;
	KEP.UserDatRef=&UserDatRef;
	KEP.Key=Key;
	KEP.File=File;
	P_pds_file GF=&_Files->Files[File->Index];
	if((Key>0)&&(Key<=GF->datInfo.NrOfKeys))
	{
    	KEP.DI=&GF->InxDescr[Key];
		if(!KESearch(KEP.DI->RootRef,&KEP))
		{
			return 0;
		}
	}
	else
	{
		File->MyFiles->LastErrorCode=err_BadKey;
		return 0;
	}
	return 1;
}

int IsamFindKey(P_local_file File,unsigned short Key,int * UserDatRef,IsamKeyStr * UserKey)
{
    *UserDatRef=0;
    int LocI=_IsamFindKey(File,Key,UserDatRef,UserKey);
    if((!LocI)&&(!File->MyFiles->LastErrorCode)&&(_Files->Files[File->Index].InxDescr[Key].AllowDupKeys))
    {
        ValidateIndex(File,Key);
        IsamKeyStr TempKey;
        LocI=IsamNextKey(File,Key,UserDatRef,&TempKey);
        if((LocI)&&(CmpKeyStr(*UserKey,TempKey)!=0))
            LocI=0;
    }
    if(LocI==0)
    {
        if((File->MyFiles->LastErrorCode==0)||((int)(File->MyFiles->LastErrorCode/100)==102))
        {
            File->MyFiles->LastErrorCode=err_KeyNotFound;
        }
    }
    return LocI;
}
