/*****************************************************************************/
/*                                                                           */
/* PDS PrunierDataServer   Copyright (c) 2009-2013 Milan Sliva /sliva@fi.cz/ */
/*                                                                           */
/* Licence: GPL 2.0                       Project HomePage: http://pds.fi.cz */
/*                                                                           */
/*****************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <signal.h>
#include <unistd.h>
#include <syslog.h>
#include <execinfo.h>

#include "messages.h"
#include "childs.h"

static void ExitApp(void);
static void ExitHandler(int sig);
static void SegFaultHandler(int sig);
static void ReloadHandler(int sig);
void OpenLog();

int server_sockfd=0;
int http_sockfd=0;

int Doned=0;
int ExitSignal=0;

int main()
{
    tzset();

    time(&start_time);

    Version.major=version_major;
    Version.minor=version_minor;
    Version.revision=version_revision;
    printf("PDS - PrunierDataServer version %d.%d.%d (compilation %s %s)   Copyright (c) 2009-2013 Milan Sliva /sliva@fi.cz/\n\n",Version.major,Version.minor,Version.revision,compilation_date,compilation_time);fflush(NULL);
    syslog(0,"PDS - PrunierDataServer version %d.%d.%d (compilation %s %s)   Copyright (c) 2009-2013 Milan Sliva /sliva@fi.cz/\n\n",Version.major,Version.minor,Version.revision,compilation_date,compilation_time);

	struct sigaction signal1;
	memset(&signal1,0,sizeof(signal1));
	struct sigaction signal2;
	memset(&signal2,0,sizeof(signal2));
	struct sigaction signal3;
	memset(&signal3,0,sizeof(signal3));
	struct sigaction signal4;
	memset(&signal4,0,sizeof(signal4));

	signal1.sa_handler=ExitHandler;
	sigemptyset(&signal1.sa_mask);
	signal1.sa_flags=0;

	signal2.sa_handler=SIG_IGN;
	sigemptyset(&signal2.sa_mask);
	signal2.sa_flags=SA_RESTART;

	signal3.sa_handler=SegFaultHandler;
	sigemptyset(&signal3.sa_mask);
	signal3.sa_flags=0;

	signal4.sa_handler=ReloadHandler;
	sigemptyset(&signal4.sa_mask);
    sigaddset(&signal4.sa_mask,SIGHUP);
	signal4.sa_flags=SA_RESTART;

	if((sigaction(SIGINT,&signal1,NULL)==-1)||
		(sigaction(SIGTERM,&signal1,NULL)==-1)||
		(sigaction(SIGABRT,&signal1,NULL)==-1)||
		(sigaction(SIGQUIT,&signal1,NULL)==-1)||
		(sigaction(SIGUSR1,&signal1,NULL)==-1)||
		(sigaction(SIGCHLD,&signal2,NULL)==-1)||
		(sigaction(SIGSEGV,&signal3,NULL)==-1)||
		(sigaction(SIGHUP,&signal4,NULL)==-1)
	)
	{
		message(0,"Nelze nastavit obsluhu signalu\n");
	}

	if(atexit(ExitApp)!=0)
	{
		message(0,"Nelze zaregistrovat ExitApp()\n");
		return -1;
	}

	if((InitSemaphores()!=0)&&(InitChilds()!=0))
		{
            OpenLog();

			struct sockaddr_in server_address;

			server_sockfd=socket(AF_INET,SOCK_STREAM,0);
			http_sockfd=socket(AF_INET,SOCK_STREAM,0);

			server_address.sin_family = AF_INET;
			server_address.sin_addr.s_addr = htonl(INADDR_ANY);
			server_address.sin_port = htons(Params->port);
			int server_len = sizeof(server_address);
			P_pds_child free_child;
			status_message msg;

			if(bind(server_sockfd,(struct sockaddr *)&server_address,server_len)==0)
			{
				if(listen(server_sockfd,SOMAXCONN)==0)
				{
                    int LocI=1;
                    setsockopt(server_sockfd,SOL_TCP,TCP_NODELAY,&LocI,sizeof(LocI));
                    LocI=0;
                    setsockopt(server_sockfd,SOL_TCP,TCP_CORK,&LocI,sizeof(LocI));
					free_child=AllocChild();
					if(free_child!=NULL)
					{
					/*
						memset(&msg,0,sizeof(msg));
						msg.message_type=free_child->pid;
						msg.sign=valid_sign;
						msg.status=child_freezepatrol;
						if(msgsnd(msg_id,&msg,status_message_size,0)==0)
						{
							message(1,"FreezePatrol inicializovan ...\n");
						}
						else
						{
							message(0,"Nelze inicializovat FreezePatrol ...\n");
						}
						free_child=AllocChild();
						if(free_child!=NULL) */
                        if(Params->Echo)
                        {
                            memset(&msg,0,sizeof(msg));
                            msg.message_type=free_child->pid;
                            msg.sign=valid_sign;
                            msg.status=child_echoserver;
                            if(msgsnd(msg_id,&msg,status_message_size,0)==0)
                            {
                                message(2,"Echoserver inicializovan ...\n");
                            }
                            else
                            {
                                message(0,"Nelze inicializovat EchoServer ...\n");
                            }
                            free_child=AllocChild();
                        }
                        else
                        {
                            message(2,"Echoserver disabled ...\n");
                        }
						if(free_child!=NULL)
						{
                            if(Params->Http)
                            {
                                server_address.sin_family = AF_INET;
                                server_address.sin_addr.s_addr = htonl(INADDR_ANY);
                                server_address.sin_port = htons(Params->httpport);
                                if(bind(http_sockfd,(struct sockaddr *)&server_address,server_len)==0)
                                {
                                    if(listen(http_sockfd,SOMAXCONN)==0)
                                    {
                                        LocI=1;
                                        setsockopt(http_sockfd,SOL_TCP,TCP_NODELAY,&LocI,sizeof(LocI));
                                        LocI=0;
                                        setsockopt(http_sockfd,SOL_TCP,TCP_CORK,&LocI,sizeof(LocI));
                                        memset(&msg,0,sizeof(msg));
                                        msg.message_type=free_child->pid;
                                        msg.sign=valid_sign;
                                        msg.status=child_httpserver;
                                        msg.data=http_sockfd;
                                        if(msgsnd(msg_id,&msg,status_message_size,0)==0)
                                        {
                                            message(2,"HTTP server inicializovan ...\n");
                                        }
                                        else
                                        {
                                            message(0,"Nelze inicializovat HTTP server ...\n");
                                        }
                                        free_child=AllocChild();
                                    }
                                }
                            }
						}
                        else
                        {
                            message(2,"HTTP server disabled ...\n");
                        }
						if(free_child!=NULL)
						{
							memset(&msg,0,sizeof(msg));
							msg.message_type=free_child->pid;
							msg.sign=valid_sign;
							msg.status=child_server;
							msg.data=server_sockfd;
							if(msgsnd(msg_id,&msg,status_message_size,0)==0)
							{
								int res_msg=0;
								status_message statusmsg;
								while((res_msg!=-1)&&(Live))
								{
									res_msg=msgrcv(msg_id,&statusmsg,status_message_size,parent_pid,0);
									if(res_msg!=-1)
									{
										WorkMsg(statusmsg);
									}
									else
									{
									    if(errno==4)
                                            res_msg=0;
									}
								}
							}
						}
					}
				}
				else
				{
					message(0,"'listen' selhal ...\n");
				}
			}
			else
			{
				message(0,"'bind' selhal ...\n");
			}
		}
	return 0;
}

void ExitApp(void)
{
	if(IAmChild())
	{
	    if(_Files!=NULL)
	    {
            if(--_Files->UseCount==0)
                DoneFiles();
	    }
		message(3,"%d Ukoncovaci procedura vlakna\n",getpid());
		CancelThreadSemaphores(getpid());
	}
	else
	if((!Doned)&&(!Live))
	{
		Doned=1;
		syslog(0,"PDS exit signal: %d\n",ExitSignal);
		message(3,"%d Ukoncovaci procedura materskeho procesu\n",getpid());
		DoneChilds();
		if(http_sockfd!=0)
		{
            close(http_sockfd);
			http_sockfd=0;
		}
		if(server_sockfd!=0)
		{
            close(server_sockfd);
			server_sockfd=0;
		}
		DoneParams();
		DoneFiles();
		DoneSemaphores();
	}
}

void ExitHandler(int sig)
{
	if(sig==SIGUSR1)
	{
		status_message msg;
		memset(&msg,0,sizeof(msg));
		msg.message_type=parent_pid;
		msg.sign=valid_sign;
		msg.status=child_abort;
		msg.pid=getpid();
		if(msgsnd(msg_id,&msg,status_message_size,0)==0)
		{
			message(3,"%d odeslana zprava o nutnosti ukoncit vlakno\n",getpid());
			exit(EXIT_FAILURE);
		}
		else
		{
			message(3,"%d nelze odeslat zpravu o nutnosti ukoncit vlakno\n",getpid());
		}
	}
	else
	{
	    ExitSignal=sig;
		message(0,"%d SIGNAL: %d\n",getpid(),sig);
		ThreadsLive=0;Live=0;
		close(server_sockfd);
		server_sockfd=0;
		exit(sig);
	}
}

static void SegFaultHandler(int sig)
{
  void *array[16];
  size_t size;

  // get void*'s for all entries on the stack
  size = backtrace(array, 16);

  // print out all the frames to stderr
  fprintf(stdout, "Error: signal %d:\n", sig);
  backtrace_symbols_fd(array, size, 1);
  exit(1);
}

static void ReloadHandler(int sig)
{
	if(IAmChild())
	{
	    kill(parent_pid,SIGHUP);
	}
	else
	{
        status_message msg;
        memset(&msg,0,sizeof(msg));
        msg.message_type=parent_pid;
        msg.sign=valid_sign;
        msg.status=Reload;
        msg.pid=getpid();
        msgsnd(msg_id,&msg,status_message_size,0);
	}
}
