#ifndef messages_h
	#define messages_h

#if defined(_WIN32) || defined(_WIN64)
    #include "wincompat.h"
#endif
    #include <stdio.h>
    #include <sys/types.h>
    #include <unistd.h>
    #include <stdlib.h>
    #include <stdint.h>
    #include <stdarg.h>
#ifdef __linux__
    #include <syslog.h>
#endif // __linux__
    #include <time.h>

    #define MaxKeyLen					96 // nesmi byt vetsi nez 255

	#define child_create 		1
	#define child_exit   		2
	#define child_free   		3
	#define child_server 		4
	#define child_allocfree		5
	#define child_returnfree	6

	#define file_appropriated	7
	#define file_next_expectant	8
	#define file_fatalerror    	9
	#define file_setexpectant  	10
	#define file_expectantok	11
	#define child_freezepatrol	12
	#define child_abort			13
	#define child_echoserver    14
	#define child_httpserver    15
	#define Reload              16

    #define allow_pds           1
    #define allow_http          2

    #define pds_Debug                   0

    #define pds_OpenFile				1
    #define pds_CloseFile				2
    #define pds_ReadRec					3
    #define pds_GetRec					4
    #define pds_ClearKey				5
    #define pds_NextKey					6
    #define pds_PrevKey					7
    #define	pds_FindKey					8
    #define pds_FindKeyAndRef   		9
    #define pds_SearchKey				10
    #define pds_SearchKeyAndRef			11
    #define pds_UsedRecs				12
    #define pds_GetBlock				13
    #define pds_PutBlock				14
    #define pds_Lock					15
    #define pds_Unlock					16
    #define pds_LockRec					17
    #define pds_UnlockRec				18
    #define pds_RecIsLocked				19
    #define pds_UnlockAllRecs			20
    #define pds_IsOnlyMyWS				21
    #define pds_OpenRawFile     		22
    #define pds_ReadLock				23
    #define pds_UsedKeys				24
    #define pds_SeekEOF					25
    #define pds_CreateFile				26
    #define pds_IAmLocker				27
    #define pds_IAmReadLocker			28
    #define pds_TotalRecs				29
    #define pds_FreeRecs				30
    #define pds_AddRec					31
    #define pds_IndexCount      		32
    #define pds_AddKey					33
    #define pds_Semaphore				34
    #define pds_GetRelPos				35
    #define pds_GetKeyInRelPos			36
    #define pds_Message					37
    #define pds_NextDiffKey				38
    #define pds_PrevDiffKey				39
    #define pds_PutRec					40
    #define pds_DeleteKey				41
    #define pds_DeleteRec				42
    #define pds_ResetIndex				43
    #define pds_GetIndexInfo			44
    #define pds_GetFromIni				45
    #define pds_SetToIni				46
    #define pds_ExistInIni				47
    #define pds_GetIniSections			48
    #define pds_GetIniKeys				49
    #define pds_DeleteIniKey			50
    #define pds_DeleteIniSection		51
    #define pds_KeyExists				52
    #define pds_ExistFile				53
    #define pds_FileSize				54
    #define pds_Identify				55
    #define pds_ChangeStructure     	56
    #define pds_XXXXXXXXXXXXXXXXXXXXXXX 57
    #define pds_ResetFile				58
    #define pds_GetFileInfo				59
    #define pds_CreateSubKey            60
    #define pds_AddRecord               61
    #define pds_MultiAddKey             62
    #define pds_ModifyRecord            63
    #define pds_DeleteRecord            64
    #define pds_CreateArchive           65
    #define pds_ExtractArchive          66
    #define pds_File                    67
    #define pds_Authentication          68
    #define pds_CloseAllFiles           69
    #define pds_GetVersion              70
    #define pds_RebuildFile             71
    #define pds_ChangeFile              72
    #define pds_GetTime                 73
    #define pds_CloneFile               74

	typedef struct status_message status_message,*P_status_message;

	struct status_message {
		long int 			message_type;
		int					sign;
		unsigned short int 	status;
		pid_t 				pid;
		int 				data;
	} ;

    struct VersionRec {
        uint16_t       major;
        uint16_t       minor;
        uint32_t       revision;
    };

    extern struct VersionRec Version;

    extern const int valid_sign;
    extern int msg_id;
    extern int Live;
    extern int ThreadsLive;
	extern const int status_message_size;
	extern int DebugLevel;
    extern uint8_t LogMode;

/*
    DebugLevel 0 - pouze chybove zpravy
               1 - i zpravy tykajici se prihlasovani, odhlasovani, vytvareni a obnovovani souboru
               2 - i zpravy tykajici se otevirani a zavirani souboru
               3 - i zpravy tykajici se vlaken
               4 - i vypis stavu souboru
               5 - i zpravy o vyhrazovani/uvolnovani souboru
               7 - i vypisy bufferu pokud se nekde vypisuji
*/

/*******/

#pragma pack(push,1)

typedef struct ResultMsg ResultMsg,*P_ResultMsg;

struct ResultMsg {
	unsigned				counter;
	unsigned short			Size;
	unsigned short  		ResultCode;
	unsigned short          Flags;  /* bits 0 nastaven 0x0001 - obsahuje Preload data */
                                    /*      1 nastaven 0x0002 - jedna se o Wait zpravu */
                                    /*      2 nastaven 0x0004 - ResultCode obsahuje u Wait zpravy progress v procentech * 100, tj. 100%=10000 */
                                    /*      3 nastaven 0x0008 - jedna se o zpravu obsahujici fragment souboru */
                                    /*      4 nastaven 0x0010 - jedna se o zpravu obsahujici MD5 soucet zaslaneho souboru */
};

struct fileID {
	uint16_t	Random;  // nahodna hodnota
	uint16_t	FileGen; // pri otevreni se dosazuje hodnota z Generation++
	uint32_t	Index;
};

typedef struct IsamKeyStr IsamKeyStr,*P_IsamKeyStr;

struct IsamKeyStr {
	uint8_t Length;
	char	Data[MaxKeyLen];
};

typedef struct FileInfoRec FileInfoRec,*P_FileInfoRec;
struct FileInfoRec {
	uint16_t	SizeOf;
	uint64_t	Size;
	uint32_t	Time;
	uint32_t	MSTime;
};

typedef struct message_head message_head,*P_message_head;

struct message_head {
	unsigned short int 	    message_id;
	unsigned short int	    message_size;
	unsigned int 			counter;
} ;

typedef struct child_message child_message,*P_child_message;

struct child_message {
	struct message_head	head;
	char 					message[];
} ;

typedef struct ResultMsg1 ResultMsg1,*P_ResultMsg1;

struct ResultMsg1 {
	struct ResultMsg 	Head;
	uint8_t				Result;
};

typedef struct ResultMsg2 ResultMsg2,*P_ResultMsg2;

struct ResultMsg2 {
	struct ResultMsg 	Head;
	struct fileID		ID;
	unsigned short		DataSize;
};

typedef struct ResultMsg3 ResultMsg3,*P_ResultMsg3;

struct ResultMsg3 {
	struct ResultMsg 	Head;
	int					Result;
};

typedef struct ResultMsg4 ResultMsg4,*P_ResultMsg4;

struct ResultMsg4 {
	struct ResultMsg 	Head;
	FileInfoRec			FileInfo;
};

typedef struct ResultInxMsg1 ResultInxMsg1,*P_ResultInxMsg1;

struct ResultInxMsg1 {
	struct ResultMsg 	Head;
	int 				RefNr;
	IsamKeyStr			KeyStr;
};

typedef struct ResultInxMsg2 ResultInxMsg2,*P_ResultInxMsg2;

struct ResultInxMsg2 {
	struct ResultMsg 	Head;
	int 				RefNr;
};

typedef struct ResultInxMsg3 ResultInxMsg3,*P_ResultInxMsg3;

struct ResultInxMsg3 {
	struct ResultMsg 	Head;
	uint16_t			WResult;
};

typedef struct ResultInxMsg4 ResultInxMsg4,*P_ResultInxMsg4;

struct ResultInxMsg4 {
	struct ResultMsg 	Head;
	uint8_t				KeyLen;
	uint8_t				AllowDupKeys;
	unsigned short		KeyCodePage;
	unsigned short		IndexVersion;
};

typedef struct FilerMsg FilerMsg,*P_FilerMsg;

struct FilerMsg {
	struct fileID	ID;
};

typedef struct FilerDatMsg1 FilerDatMsg1,*P_FilerDatMsg1;

struct FilerDatMsg1 {
	struct fileID	ID;
	int 			RefNr;
};

typedef struct FilerDatMsg2 FilerDatMsg2,*P_FilerDatMsg2;

struct FilerDatMsg2 {
	struct fileID	ID;
	int 			Offset;
	unsigned short	Len;
};

typedef struct FilerDatMsg3 FilerDatMsg3,*P_FilerDatMsg3;

struct FilerDatMsg3 {
	struct fileID	ID;
	uint8_t Byte1;
	uint8_t Byte2;
};

typedef struct FilerDatMsg4 FilerDatMsg4,*P_FilerDatMsg4;

struct FilerDatMsg4 {
	struct fileID	ID;
	unsigned short	NewLen;
};

typedef struct FilerDatMsg5 FilerDatMsg5,*P_FilerDatMsg5;

struct FilerDatMsg5 {
	struct fileID	ID;
	uint8_t	        Type;
    unsigned char   MD5[16];
};

typedef struct FilerInxMsg1 FilerInxMsg1,*P_FilerInxMsg1;

struct FilerInxMsg1 {
	struct fileID	ID;
	unsigned short	KeyNr;
};

typedef struct FilerInxMsg2 FilerInxMsg2,*P_FilerInxMsg2;

struct FilerInxMsg2 {
	struct fileID	ID;
	unsigned short	KeyNr;
	uint8_t			KeyLen;
	char			Key[];
};

typedef struct FilerInxMsg3 FilerInxMsg3,*P_FilerInxMsg3;

struct FilerInxMsg3 {
	struct fileID	ID;
	int				RefNr;
	unsigned short	KeyNr;
	short			Direction;
	uint8_t			KeyLen;
	char			Key;
};

typedef struct FilerInxMsg4 FilerInxMsg4,*P_FilerInxMsg4;

struct FilerInxMsg4 {
	struct fileID	ID;
	int 			RefNr;
	unsigned short	KeyNr;
	uint8_t			KeyLen;
	char			Key;
};

typedef struct FilerInxMsg5 FilerInxMsg5,*P_FilerInxMsg5;

struct FilerInxMsg5 {
	struct fileID	ID;
	int 			RefNr;
	unsigned short	KeyNr;
	unsigned short	Scale;
	uint8_t			KeyLen;
	char			Key;
};

typedef struct FilerInxMsg6 FilerInxMsg6,*P_FilerInxMsg6;

struct FilerInxMsg6 {
	struct fileID	ID;
	unsigned short	KeyNr;
	unsigned short	RelPos;
	unsigned short	Scale;
};

typedef struct FilerInxMsg7 FilerInxMsg7,*P_FilerInxMsg7;

struct FilerInxMsg7 {
	struct fileID	ID;
	int 			RefNr;
	unsigned short	KeyNr;
};

typedef struct MultiAddKey MultiAddKey,*P_MultiAddKey;

struct MultiAddKey {
	uint16_t        KeyNr;
	IsamKeyStr      Key;
};

typedef struct VersionRec VersionRec,*P_VersionRec;

typedef struct TransferFileMsg TransferFileMsg,*P_TransferFileMsg;

struct TransferFileMsg {
	int                  FilePos;
};

typedef struct TransferFileLastMsg TransferFileLastMsg,*P_TransferFileLastMsg;

struct TransferFileLastMsg {
	int                  FilePos;       /* -1 */
	int                  FileSize;
	unsigned char        MD5Digest[16];
};

typedef struct KeyInfoRec KeyInfoRec,*P_KeyInfoRec;
struct KeyInfoRec {
	uint8_t				KeyLen;
	uint8_t				AllowDupKeys;
	unsigned short		KeyCodePage;
	unsigned short		IndexVersion;
	int32_t             UsedKeys;
};

typedef struct VersionRecMsg VersionRecMsg,*P_VersionRecMsg;
struct VersionRecMsg {
	struct ResultMsg 	Head;
	uint16_t			Major;
	uint16_t			Minor;
	uint32_t			Revision;
};

typedef struct TimeRecMsg TimeRecMsg,*P_TimeRecMsg;
struct TimeRecMsg {
	struct ResultMsg 	Head;
	uint64_t			TimeStamp;
	uint16_t            year;
    uint8_t             month;
    uint8_t             day;
    uint8_t             hour;
    uint8_t             min;
    uint8_t             sec;
    uint8_t             wday;
    uint16_t            yday;
    uint8_t             isdst;
    int16_t             TZMinutesOffset;
    char                TimeZoneName[16];
};

#pragma pack(pop)

/*******/

    void message(int level,const char* format, ... );
    int StrTime(time_t Time,char * Buffer,int BufferSize);
    int StrUptime(time_t Time,char * Buffer,int BufferSize);

#endif
