#include "semaphores.h"

int sem_id=-1;
//int ChildFlag=0;
pid_t parent_pid=0;
//uid_t uid=0xFFFF;
//uid_t gid=0xFFFF;
uint8_t UseSemaphores=0;

typedef struct local_sem local_sem,*P_local_sem;

struct local_sem {
	pid_t			pid;
	int				counter;
};

local_sem LSem[SEMMSL];

int MaxIndexSemaphores;

#ifndef LOCAL

int semaphore_p(int sem_num,unsigned short * ErrAddr)
{
    if(UseSemaphores)
    {
        if(sem_num>1)
        {
            if(LSem[sem_num].pid==getpid())
            {
                LSem[sem_num].counter++;
                return(1);
            }
        }

        struct sembuf sem_b;

        sem_b.sem_num = sem_num;
        sem_b.sem_op = -1; /* P() */
        sem_b.sem_flg = SEM_UNDO;
        if (semop(sem_id, &sem_b, 1) == -1) {
            message(0,"%d semaphore_p %d failed\n",getpid(),sem_num);
            fprintf(stderr, "%d semaphore_p %d failed\n",getpid(),sem_num);
            if(ErrAddr!=NULL)
            {
                *ErrAddr=err_SemaphoreFailed;
            }
            return(0);
        }

        LSem[sem_num].counter=1;
        LSem[sem_num].pid=getpid();
    }
    return(1);
}

int semaphore_v(int sem_num,unsigned short * ErrAddr)
{
    if(UseSemaphores)
    {
        if(sem_num>1)
        {
            if(LSem[sem_num].pid==getpid())
            {
                if(LSem[sem_num].counter)
                {
                    LSem[sem_num].counter--;
                }
                else
                {
                    message(0,"%d Detekovano chybne uvolneni semaforu %d\n",getpid(),sem_num);
                }
                if(LSem[sem_num].counter)
                    return(1);
            }
        }

        LSem[sem_num].counter=0;
        LSem[sem_num].pid=0;

        struct sembuf sem_b;

        sem_b.sem_num = sem_num;
        sem_b.sem_op = 1; /* V() */
        sem_b.sem_flg = SEM_UNDO;
        if (semop(sem_id, &sem_b, 1) == -1) {
            message(0,"%d semaphore_v %d failed\n",getpid(),sem_num);
            fprintf(stderr, "%d semaphore_v %d failed\n",getpid(),sem_num);
            if(ErrAddr!=NULL)
            {
                *ErrAddr=err_SemaphoreFailed;
            }
            return(0);
        }
    }
    return(1);
}

int set_semvalue(int sem_num)
{
    union semun sem_union;
    sem_union.val = 1;
    if (semctl(sem_id,sem_num,SETVAL,sem_union) == -1) return(0);
    return(1);
}

void del_semvalue()
{
/*    union semun sem_union;
	memset(&sem_union,0,sizeof(sem_union));*/
    if(sem_id!=-1)
	{
    	if (semctl(sem_id,0,IPC_RMID/*,sem_union*/) == -1)
        	message(0,"Uvolneni semaforu selhalo\n");
        	fprintf(stderr, "Uvolneni semaforu selhalo\n");
	}
}

int InitSemaphores()
{
	memset(&LSem,0,sizeof(LSem));

	sem_id=semget(IPC_PRIVATE,SEMMSL,666|IPC_CREAT);
	if(sem_id==-1)
    {
		message(0,"Chyba vytvoreni semaforu ...\n");
		printf("Chyba vytvoreni semaforu ...\n");
		return 0;
	}
	MaxIndexSemaphores=SEMMSL-MaxSemaphores;
	int i;
	for(i=0;i<SEMMSL;i++)
	{
		if(!set_semvalue(i))
  		{
            message(0,"Chyba vytvoreni semaforu ...\n");
            printf("Chyba vytvoreni semaforu ...\n");
	   		return 0;
  		}
  	}
  	UseSemaphores=1;
	return 1;
}

void DoneSemaphores()
{
	if(!IAmChild())
	{
	    UseSemaphores=0;
		del_semvalue();
	}
}

void CancelThreadSemaphores(pid_t thread_pid)
{
	int I=0;
	for(I=0;I<SEMMSL;I++)
	{
		if(LSem[I].pid==thread_pid)
		{
			LSem[I].counter=1;
			semaphore_v(I,NULL);
		}
	}
}

#else

static wxCriticalSection * Critical[SEMMSL]={NULL};

int InitSemaphores()
{
	MaxIndexSemaphores=SEMMSL-MaxSemaphores;
	int LocI=0;
	while(LocI<SEMMSL)
    {
        if(Critical[LocI]==NULL) Critical[LocI]=new wxCriticalSection();
        LocI++;
    }
  	UseSemaphores=1;
	return 1;
}

int semaphore_p(int sem_num,unsigned short * ErrAddr)
{
    if(UseSemaphores)
    {
        if(sem_num>SEMMSL-1)
            return(0);
        if(Critical[sem_num]==NULL) return(0);
        if(sem_num>1)
        {
            if(LSem[sem_num].pid==getpid())
            {
                LSem[sem_num].counter++;
                return(1);
            }
        }

        Critical[sem_num]->Enter();

        LSem[sem_num].counter=1;
        LSem[sem_num].pid=getpid();
    }
    return(1);
}

int semaphore_v(int sem_num,unsigned short * ErrAddr)
{
    if(UseSemaphores)
    {
        if(sem_num>SEMMSL-1)
            return(0);
        if(Critical[sem_num]==NULL) return(0);
        if(sem_num>1)
        {
            if(LSem[sem_num].pid==getpid())
            {
                if(LSem[sem_num].counter)
                {
                    LSem[sem_num].counter--;
                }
                else
                {
                    message(0,"%d Detekovano chybne uvolneni semaforu %d\n",getpid(),sem_num);
                }
                if(LSem[sem_num].counter)
                    return(1);
            }
        }

        LSem[sem_num].counter=0;
        LSem[sem_num].pid=0;

        Critical[sem_num]->Leave();
    }
    return(1);
}

void DoneSemaphores()
{
	if(!IAmChild())
	{
	    UseSemaphores=0;
        int LocI=0;
        while(LocI<SEMMSL)
        {
            if(Critical[LocI]!=NULL)
            {
                delete Critical[LocI];
                Critical[LocI]=NULL;
            }
            LocI++;
        }
	}
}

#endif

int index_semaphore_p(int Index)
{
	return semaphore_p(MaxSemaphores+Index%MaxIndexSemaphores,NULL);
}

int index_semaphore_v(int Index)
{
	return semaphore_v(MaxSemaphores+Index%MaxIndexSemaphores,NULL);
}

void * GetMem(size_t size)
{
    if(!semaphore_p(semaphore_malloc,NULL))
      return NULL;
    void * allocated=malloc(size);
    semaphore_v(semaphore_malloc,NULL);
    return allocated;
}

void FreeMem(void * ptr)
{
    if(semaphore_p(semaphore_malloc,NULL))
    {
        free(ptr);
        semaphore_v(semaphore_malloc,NULL);
    }
}

void * ReallocMem(void * ptr,size_t newsize)
{
    void * Result=NULL;
    if(semaphore_p(semaphore_malloc,NULL))
    {
        Result=realloc(ptr,newsize);
        semaphore_v(semaphore_malloc,NULL);
    }
    return Result;
}

int IAmChild()
{
	if(getpid()==parent_pid)
		return 0;
	else
		return 1;
}
